from __future__ import annotations

from typing import Any

from checkov.common.models.enums import CheckCategories
from checkov.terraform.checks.resource.base_resource_value_check import BaseResourceValueCheck


class TransferServerIsPublic(BaseResourceValueCheck):
    def __init__(self) -> None:
        name = "Ensure Transfer Server is not exposed publicly."
        id = "CKV_AWS_164"
        supported_resources = ("aws_transfer_server",)
        categories = (CheckCategories.GENERAL_SECURITY,)
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def get_inspected_key(self) -> str:
        return "endpoint_type"

    def get_expected_values(self) -> list[Any]:
        return ["VPC", "VPC_ENDPOINT"]


check = TransferServerIsPublic()



from __future__ import annotations

import itertools
import json
import logging
import os
from collections import defaultdict
from collections.abc import Sequence
from copy import deepcopy
from pathlib import Path
from typing import Any, Optional, TYPE_CHECKING, TypeVar, cast

import hcl2
from lark import Tree
import re

from checkov.common.util.consts import DEFAULT_EXTERNAL_MODULES_DIR
from checkov.common.util.json_utils import CustomJSONEncoder, object_hook
from checkov.terraform.checks.utils.dependency_path_handler import unify_dependency_path
from checkov.terraform.graph_builder.utils import remove_module_dependency_in_path
from checkov.common.util.parser_utils import TERRAFORM_NESTED_MODULE_PATH_PREFIX, TERRAFORM_NESTED_MODULE_PATH_ENDING, \
    is_nested, get_abs_path, get_module_from_full_path

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

_Conf = TypeVar("_Conf", bound="dict[Any, Any]")

ENTITY_NAME_PATTERN = re.compile(r"[^\W0-9][\w-]*")
RESOLVED_MODULE_PATTERN = re.compile(r"\[.+\#.+\]")
_Hcl2Payload: TypeAlias = "dict[str, list[dict[str, Any]]]"
external_modules_download_path = os.environ.get('EXTERNAL_MODULES_DIR', DEFAULT_EXTERNAL_MODULES_DIR)


def is_valid_block(block: Any) -> bool:
    if not isinstance(block, dict):
        return True

    # if the block is empty, there's no need to process it further
    if not block:
        return False

    entity_name = next(iter(block.keys()))
    if re.fullmatch(ENTITY_NAME_PATTERN, entity_name):
        return True
    return False


def validate_malformed_definitions(raw_data: _Hcl2Payload) -> _Hcl2Payload:
    return {
        block_type: [block for block in blocks if is_valid_block(block)]
        for block_type, blocks in raw_data.items()
    }


def load_or_die_quietly(
    file: str | Path | os.DirEntry[str], parsing_errors: dict[str, Exception], clean_definitions: bool = True
) -> Optional[_Hcl2Payload]:
    """
Load JSON or HCL, depending on filename.
    :return: None if the file can't be loaded
    """

    file_path = os.fspath(file)
    file_name = os.path.basename(file_path)

    try:
        logging.debug(f"Parsing {file_path}")

        with open(file_path, "r", encoding="utf-8-sig") as f:
            if file_name.endswith(".json"):
                return cast("_Hcl2Payload", json.load(f))
            else:
                raw_data = hcl2.load(f)
                non_malformed_definitions = validate_malformed_definitions(raw_data)
                if clean_definitions:
                    return clean_bad_definitions(non_malformed_definitions)
                else:
                    return non_malformed_definitions
    except Exception as e:
        logging.debug(f'failed while parsing file {file_path}', exc_info=True)
        parsing_errors[file_path] = e
        return None


def clean_bad_definitions(tf_definition_list: _Hcl2Payload) -> _Hcl2Payload:
    return {
        block_type: [
            definition
            for definition in definition_list
            if block_type in {"locals", "terraform"} or not isinstance(definition, dict) or len(definition) == 1
        ]
        for block_type, definition_list in tf_definition_list.items()
    }


def safe_index(sequence_hopefully: Sequence[Any], index: int) -> Any:
    try:
        return sequence_hopefully[index]
    except IndexError:
        logging.debug(f'Failed to parse index int ({index}) out of {sequence_hopefully}', exc_info=True)
        return None


def remove_module_dependency_from_path(path: str) -> str:
    """
    :param path: path that looks like "dir/main.tf[other_dir/x.tf#0]
    :return: only the outer path: dir/main.tf
    """
    if re.findall(RESOLVED_MODULE_PATTERN, path):
        path = re.sub(RESOLVED_MODULE_PATTERN, '', path)
    return path


def get_module_dependency_map(
    tf_definitions: dict[str, Any]
) -> tuple[dict[str, list[list[str]]], dict[str, Any], dict[tuple[str, str], list[str]]]:
    """
    :param tf_definitions, with paths in format 'dir/main.tf[module_dir/main.tf#0]'
    :return module_dependency_map: mapping between directories and the location of its module definition:
            {'dir': 'module_dir/main.tf'}
    :return tf_definitions: with paths in format 'dir/main.tf'
    """
    module_dependency_map: dict[str, list[list[str]]] = {}
    copy_of_tf_definitions = {}
    dep_index_mapping: dict[tuple[str, str], list[str]] = {}
    origin_keys = list(filter(lambda k: not k.endswith(TERRAFORM_NESTED_MODULE_PATH_ENDING), tf_definitions.keys()))
    unevaluated_keys = list(filter(lambda k: k.endswith(TERRAFORM_NESTED_MODULE_PATH_ENDING), tf_definitions.keys()))
    for file_path in origin_keys:
        dir_name = os.path.dirname(file_path)
        module_dependency_map[dir_name] = [[]]
        copy_of_tf_definitions[file_path] = deepcopy(tf_definitions[file_path])

    next_level, unevaluated_keys = get_next_vertices(origin_keys, unevaluated_keys)
    while next_level:
        for file_path in next_level:
            path, module_dependency, module_dependency_num = remove_module_dependency_in_path(file_path)
            dir_name = os.path.dirname(path)
            current_deps = deepcopy(module_dependency_map[os.path.dirname(module_dependency)])
            for dep in current_deps:
                dep.append(module_dependency)
            if dir_name not in module_dependency_map:
                module_dependency_map[dir_name] = current_deps
            else:
                for dep in current_deps:
                    if dep not in module_dependency_map[dir_name]:
                        module_dependency_map[dir_name].append(dep)
            copy_of_tf_definitions[path] = deepcopy(tf_definitions[file_path])
            origin_keys.append(path)
            dep_index_mapping.setdefault((path, module_dependency), []).append(module_dependency_num)
        next_level, unevaluated_keys = get_next_vertices(origin_keys, unevaluated_keys)
    for key, dep_trails in module_dependency_map.items():
        hashes = set()
        deduped = []
        for trail in dep_trails:
            trail_hash = unify_dependency_path(trail)
            if trail_hash in hashes:
                continue
            hashes.add(trail_hash)
            deduped.append(trail)
        module_dependency_map[key] = deduped
    return module_dependency_map, copy_of_tf_definitions, dep_index_mapping


def get_next_vertices(evaluated_files: list[str], unevaluated_files: list[str]) -> tuple[list[str], list[str]]:
    """
    This function implements a lazy separation of levels for the evaluated files. It receives the evaluated
    files, and returns 2 lists:
    1. The next level of files - files from the unevaluated_files which have no unresolved dependency (either
        no dependency or all dependencies were evaluated).
    2. unevaluated - files which have yet to be evaluated, and still have pending dependencies

    Let's say we have this dependency tree:
    a -> b
    x -> b
    y -> c
    z -> b
    b -> c
    c -> d

    The first run will return [a, y, x, z] as the next level since all of them have no dependencies
    The second run with the evaluated being [a, y, x, z] will return [b] as the next level.
    Please mind that [c] has some resolved dependencies (from y), but has unresolved dependencies from [b].
    The third run will return [c], and the fourth will return [d].
    """

    next_level, unevaluated, do_not_eval_yet = [], [], []
    for key in unevaluated_files:
        found = False
        for eval_key in evaluated_files:
            if eval_key in key:
                found = True
                break
        if not found:
            do_not_eval_yet.append(key.split(TERRAFORM_NESTED_MODULE_PATH_PREFIX)[0])
            unevaluated.append(key)
        else:
            next_level.append(key)

    move_to_uneval = list(filter(lambda k: k.split(TERRAFORM_NESTED_MODULE_PATH_PREFIX)[0] in do_not_eval_yet, next_level))
    for k in move_to_uneval:
        next_level.remove(k)
        unevaluated.append(k)
    return next_level, unevaluated


def get_module_dependency_map_support_nested_modules(
    tf_definitions: dict[str, Any]
) -> tuple[dict[str, list[list[str]]], dict[str, Any], dict[tuple[str, str | None], Any]]:
    module_dependency_map: dict[str, list[list[str]]] = defaultdict(list)
    dep_index_mapping = defaultdict(list)
    for tf_definition_key in tf_definitions.keys():
        if not is_nested(tf_definition_key):
            dir_name = os.path.dirname(tf_definition_key)
            module_dependency_map[dir_name].append([])
            continue
        modules_list, path = get_nested_modules_data_as_list(tf_definition_key)
        dir_name = os.path.dirname(path)
        module_dependency_map[dir_name].append([m for m, i in modules_list if m])
        dep_index_mapping[(path, modules_list[-1][0])].append(modules_list[-1][1])

    for key, dir_list in module_dependency_map.items():
        dir_list.sort()
        module_dependency_map[key] = list(dir_list for dir_list, _ in itertools.groupby(dir_list))
    return module_dependency_map, tf_definitions, dep_index_mapping


def get_nested_modules_data_as_list(file_path: str) -> tuple[list[tuple[str | None, str | None]], str]:
    path = get_abs_path(file_path)
    module_path: str | None = file_path
    modules_list = []

    while is_nested(module_path):
        module, index = get_module_from_full_path(module_path)
        modules_list.append((module, index))
        module_path = module
    modules_list.reverse()
    return modules_list, path


def clean_parser_types(conf: _Conf) -> _Conf:
    if not conf:
        return conf

    sorted_keys = list(conf.keys())
    first_key_type = type(sorted_keys[0])
    if first_key_type is None:
        return {}

    if all(isinstance(x, first_key_type) for x in sorted_keys):
        sorted_keys.sort()

    # Create a new dict where the keys are sorted alphabetically
    sorted_conf = {key: conf[key] for key in sorted_keys}
    for attribute, values in sorted_conf.items():
        if attribute == 'alias':
            continue
        if isinstance(values, list):
            sorted_conf[attribute] = clean_parser_types_lst(values)
        elif isinstance(values, dict):
            sorted_conf[attribute] = clean_parser_types(values)
        elif isinstance(values, str) and values in ('true', 'false'):
            sorted_conf[attribute] = True if values == 'true' else False
        elif isinstance(values, set):
            sorted_conf[attribute] = clean_parser_types_lst(list(values))
        elif isinstance(values, Tree):
            sorted_conf[attribute] = str(values)
    return sorted_conf  # type:ignore[return-value]  # still the same type as before


def clean_parser_types_lst(values: list[Any]) -> list[Any]:
    for idx, val in enumerate(values):
        if isinstance(val, dict):
            values[idx] = clean_parser_types(val)
        elif isinstance(val, list):
            values[idx] = clean_parser_types_lst(val)
        elif isinstance(val, str):
            if val == 'true':
                values[idx] = True
            elif val == 'false':
                values[idx] = False
        elif isinstance(val, set):
            values[idx] = clean_parser_types_lst(list(val))
    str_values_in_lst = []
    result_values = []
    for val in values:
        if isinstance(val, str):
            str_values_in_lst.append(val)
        else:
            result_values.append(val)
    str_values_in_lst.sort()
    result_values.extend(str_values_in_lst)
    return result_values


def serialize_definitions(tf_definitions: _Conf) -> _Conf:
    return cast("_Conf", json.loads(json.dumps(tf_definitions, cls=CustomJSONEncoder), object_hook=object_hook))



from __future__ import annotations

import itertools
import logging
from pathlib import Path
from typing import Any

from charset_normalizer import from_path
from yaml.scanner import ScannerError
from yaml import YAMLError

from checkov.common.parsers.json import parse as json_parse
from checkov.common.parsers.yaml import loader
from checkov.common.util.consts import LINE_FIELD_NAMES
from checkov.terraform.graph_builder.graph_components.block_types import BlockType

COMMENT_FIELD_NAME = "//"
IGNORE_FILED_NAMES = {COMMENT_FIELD_NAME} | LINE_FIELD_NAMES
SIMPLE_TYPES = (str, int, float, bool)

logger = logging.getLogger(__name__)


def parse(file_path: Path) -> tuple[dict[str, Any], list[tuple[int, str]]] | tuple[None, None]:
    """Parse file to dict object"""

    template = None
    template_lines = None
    try:
        template, template_lines = loads(file_path=file_path)
    except IOError as e:
        if e.errno == 2:
            logger.error(f"Template file not found: {file_path}")
        elif e.errno == 21:
            logger.error(f"Template references a directory, not a file: {file_path}")
        elif e.errno == 13:
            logger.error(f"Permission denied when accessing template file: {file_path}")
    except UnicodeDecodeError:
        logger.error(f"Cannot read file contents: {file_path}")
    except ScannerError as err:
        if err.problem in ("found character '\\t' that cannot start any token", "found unknown escape character"):
            try:
                result = json_parse(file_path, allow_nulls=False)
                if result:
                    template, template_lines = result  # type:ignore[assignment]  # this is handled by the next line
                    if isinstance(template, list):
                        # should not happen and is more relevant for type safety
                        template = template[0]
            except Exception:
                logger.error(f"Template {file_path} is malformed: {err.problem}")
                logger.error(f"Tried to parse {file_path} as JSON", exc_info=True)
    except YAMLError:
        pass

    if template is None or template_lines is None:
        return None, None

    return template, template_lines


def loads(file_path: Path) -> tuple[dict[str, Any], list[tuple[int, str]]]:
    """Loads the given JSON file with line numbers"""

    try:
        content = file_path.read_text()
    except UnicodeDecodeError:
        logging.debug(f"Encoding for file {file_path} is not UTF-8, trying to detect it")
        content = str(from_path(file_path).best())

    if not all(key in content for key in ("resource", "provider")):
        return {}, []

    file_lines = [(idx + 1, line) for idx, line in enumerate(content.splitlines(keepends=True))]

    template: "dict[str, Any] | list[dict[str, Any]]" = loader.loads(content=content)
    if not template:
        template = {}
    if isinstance(template, list):
        template = template[0]

    if template:
        template = prepare_definition(template)

    return template, file_lines


def prepare_definition(definition: dict[str, Any]) -> dict[str, Any]:
    definition_new: dict[str, Any] = {}

    for block_type, blocks in definition.items():
        if block_type == COMMENT_FIELD_NAME or block_type in LINE_FIELD_NAMES:
            continue

        definition_new[block_type] = []
        for block_name, config in blocks.items():
            if block_name == COMMENT_FIELD_NAME or block_name in LINE_FIELD_NAMES:
                continue

            if block_type == BlockType.RESOURCE:
                # resource have an extra nested level resource_type -> resource_name -> resource_config
                for resource_name, resource_config in config.items():
                    if resource_name in IGNORE_FILED_NAMES:
                        continue
                    definition_new[block_type].append({block_name: {resource_name: hclify(obj=resource_config)}})
            elif block_type == BlockType.PROVIDER:
                # provider are stored as a list, which we need to move one level higher to add the name
                for provider_config in config:
                    definition_new[block_type].append({block_name: hclify(obj=provider_config)})
            else:
                definition_new[block_type].append({block_name: hclify(obj=config)})

    return definition_new


def hclify(
    obj: dict[str, Any],
    conf: dict[str, Any] | None = None,
    parent_key: str | None = None,
) -> dict[str, list[Any]]:
    ret_dict = {}

    if not isinstance(obj, dict):
        raise Exception("this method receives only dicts")

    for key, value in obj.items():
        if key == COMMENT_FIELD_NAME:
            # don't hclify the comment block
            ret_dict[key] = value
        elif _is_simple_type(value) or _is_list_of_simple_types(value):
            if key in IGNORE_FILED_NAMES:
                ret_dict[key] = value
            elif parent_key == "tags":
                ret_dict[key] = value
            else:
                ret_dict[key] = _clean_simple_type_list([value])
        elif _is_list_of_dicts(value):
            child_list = []
            conf_val = conf.get(key, []) if conf else []
            if not isinstance(conf_val, list):
                # this occurs, when a resource in the current state has no value for that argument
                conf_val = [conf_val]

            for internal_val, internal_conf_val in itertools.zip_longest(value, conf_val):
                if isinstance(internal_val, dict):
                    child_list.append(hclify(internal_val, internal_conf_val, parent_key=key))
            if key == "tags":
                ret_dict[key] = [child_list]
            else:
                ret_dict[key] = child_list
        elif isinstance(value, dict):
            child_dict = hclify(value, parent_key=key)
            if parent_key == "tags":
                ret_dict[key] = child_dict
            else:
                ret_dict[key] = [child_dict]

    return ret_dict


def _is_simple_type(obj: Any) -> bool:
    if obj is None:
        return True
    if isinstance(obj, SIMPLE_TYPES):
        return True
    return False


def _is_list_of_simple_types(obj: Any) -> bool:
    if not isinstance(obj, list):
        return False
    for i in obj:
        if not _is_simple_type(i):
            return False
    return True


def _is_list_of_dicts(obj: Any) -> bool:
    if not isinstance(obj, list):
        return False
    for i in obj:
        if isinstance(i, dict):
            return True
    return False


def _clean_simple_type_list(value_list: list[Any]) -> list[Any]:
    """
    Given a list of simple types return a cleaned list of simple types.
    Converts booleans that are input as strings back to booleans to maintain consistent expectations for later evaluation.
    Sometimes Terraform Plan will output Map values as strings regardless of boolean input.
    """
    for i in range(len(value_list)):
        if isinstance(value_list[i], str):
            lower_case_value = value_list[i].lower()
            if lower_case_value == "true":
                value_list[i] = True
            if lower_case_value == "false":
                value_list[i] = False
    return value_list



from __future__ import annotations

import logging
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Type, Generator

if TYPE_CHECKING:
    from checkov.common.parsers.json.decoder import Mark


LOGGER = logging.getLogger(__name__)


class TemplateAttributeError(AttributeError):
    """ Custom error to capture Attribute Errors in the Template """


class StrNode(str):
    """Node class created based on the input class"""

    def __init__(self, x: str, start_mark: Mark | Any, end_mark: Mark | Any) -> None:
        try:
            super().__init__(x)  # type:ignore[call-arg]
        except TypeError:
            super().__init__()
        self.start_mark = start_mark
        self.end_mark = end_mark

    # pylint: disable=bad-classmethod-argument, unused-argument
    def __new__(cls, x: str, start_mark: Mark | None = None, end_mark: Mark | None = None) -> StrNode:
        return str.__new__(cls, x)

    def __getattr__(self, name: str) -> Any:
        raise TemplateAttributeError(f'{name} is invalid')

    def __deepcopy__(self, memo: dict[int, Any]) -> StrNode:
        result = StrNode(self, self.start_mark, self.end_mark)
        memo[id(self)] = result
        return result

    def __copy__(self) -> StrNode:
        return self


class DictNode(dict):  # type:ignore[type-arg]  # either typing works or runtime, but not both
    """Node class created based on the input class"""

    def __init__(self, x: dict[str, Any], start_mark: Mark | Any, end_mark: Mark | Any):
        try:
            super().__init__(x)
        except TypeError:
            super().__init__()
        self.start_mark = start_mark
        self.end_mark = end_mark
        self.condition_functions = ['Fn::If']

    def __deepcopy__(self, memo: dict[int, Any]) -> DictNode:
        result = DictNode(self, self.start_mark, self.end_mark)
        memo[id(self)] = result
        for k, v in self.items():
            result[deepcopy(k)] = deepcopy(v, memo)

        return result

    def __copy__(self) -> DictNode:
        return self

    def is_function_returning_object(self, _mappings: Any = None) -> bool:
        """
            Check if an object is using a function that could return an object
            Return True when
                Fn::Select:
                - 0  # or any number
                - !FindInMap [mapname, key, value] # or any mapname, key, value
            Otherwise False
        """
        if len(self) == 1:
            for k, v in self.items():
                if k in ['Fn::Select']:
                    if isinstance(v, list):
                        if len(v) == 2:
                            p_v = v[1]
                            if isinstance(p_v, dict):
                                if len(p_v) == 1:
                                    for l_k in p_v.keys():
                                        if l_k == 'Fn::FindInMap':
                                            return True

        return False

    def get(self, key: str, default: Any = None) -> Any:
        """ Override the default get """
        if isinstance(default, dict):
            default = DictNode(default, self.start_mark, self.end_mark)
        return super().get(key, default)

    def get_safe(
        self, key: str, default: Any = None, path: list[str] | None = None, type_t: Type[tuple[Any, ...]] = tuple
    ) -> list[tuple[tuple[Any, ...], list[str]]]:
        """Get values in format"""

        path = path or []
        value = self.get(key, default)
        if not isinstance(value, dict):
            if isinstance(value, type_t) or not type_t:
                return [(value, (path[:] + [key]))]

        results = []
        for sub_v, sub_path in value.items_safe(path + [key]):
            if isinstance(sub_v, type_t) or not type_t:
                results.append((sub_v, sub_path))

        return results

    def items_safe(
        self, path: list[int | str] | None = None, type_t: Type[tuple[Any, ...]] = tuple
    ) -> Generator[tuple[Any, ...], Any, None]:
        """Get items while handling IFs"""

        path = path or []
        if len(self) == 1:
            for k, v in self.items():
                if k == 'Fn::If':
                    if isinstance(v, list):
                        if len(v) == 3:
                            for i, if_v in enumerate(v[1:]):
                                if isinstance(if_v, DictNode):
                                    # yield from if_v.items_safe(path[:] + [k, i - 1])
                                    # Python 2.7 support
                                    for items, p in if_v.items_safe(path[:] + [k, i + 1]):
                                        if isinstance(items, type_t) or not type_t:
                                            yield items, p
                                elif isinstance(if_v, list):
                                    if isinstance(if_v, type_t) or not type_t:
                                        yield if_v, path[:] + [k, i + 1]
                                else:
                                    if isinstance(if_v, type_t) or not type_t:
                                        yield if_v, path[:] + [k, i + 1]
                elif not (k == 'Ref' and v == 'AWS::NoValue'):
                    if isinstance(self, type_t) or not type_t:
                        yield self, path[:]
        else:
            if isinstance(self, type_t) or not type_t:
                yield self, path[:]

    def __getattr__(self, name: str) -> Any:
        raise TemplateAttributeError(f'{name} is invalid')


class ListNode(list):  # type:ignore[type-arg]  # either typing works or runtime, but not both
    """Node class created based on the input class"""

    def __init__(self, x: list[Any], start_mark: Mark | Any, end_mark: Mark | Any) -> None:
        try:
            super().__init__(x)
        except TypeError:
            super().__init__()
        self.start_mark = start_mark
        self.end_mark = end_mark
        self.condition_functions = ['Fn::If']

    def __deepcopy__(self, memo: dict[int, Any]) -> ListNode:
        result = ListNode([], self.start_mark, self.end_mark)
        memo[id(self)] = result
        for v in self:
            result.append(deepcopy(v, memo))

        return result

    def __copy__(self) -> ListNode:
        return self

    def items_safe(
        self, path: list[int | str] | None = None, type_t: Type[tuple[Any, ...]] = tuple
    ) -> Generator[tuple[Any, ...], Any, None]:
        """Get items while handling IFs"""

        path = path or []
        for i, v in enumerate(self):
            if isinstance(v, DictNode):
                for items, p in v.items_safe(path[:] + [i]):
                    if isinstance(items, type_t) or not type_t:
                        yield items, p
            else:
                if isinstance(v, type_t) or not type_t:
                    yield v, path[:] + [i]

    def __getattr__(self, name: str) -> Any:
        raise TemplateAttributeError(f'{name} is invalid')



from checkov.common.models.enums import CheckCategories
from checkov.cloudformation.checks.resource.base_resource_value_check import BaseResourceValueCheck


class QLDBLedgerPermissionsMode(BaseResourceValueCheck):
    def __init__(self) -> None:
        name = "Ensure QLDB ledger permissions mode is set to STANDARD"
        id = "CKV_AWS_170"
        supported_resources = ["AWS::QLDB::Ledger"]
        categories = [CheckCategories.IAM]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def get_inspected_key(self) -> str:
        return "Properties/PermissionsMode"

    def get_expected_value(self) -> str:
        return "STANDARD"


check = QLDBLedgerPermissionsMode()



from typing import Dict, List, Any

from checkov.terraform.checks.data.base_check import BaseDataCheck
from checkov.common.models.enums import CheckResult, CheckCategories


class GooglePolicyIsPrivate(BaseDataCheck):
    def __init__(self) -> None:
        """
        ensure policy defined is not public
        policy should not set 'allUsers' or 'allAuthenticatedUsers' in the attribute 'member'/'members'
        """
        name = "Ensure IAM policy should not define public access"
        id = "CKV_GCP_113"
        supported_data = ("google_iam_policy",)
        categories = (CheckCategories.IAM,)
        super().__init__(name=name, id=id, categories=categories, supported_data=supported_data)

    def scan_data_conf(self, conf: Dict[str, List[Any]]) -> CheckResult:
        """
            validates gcp policy
        """
        fails = ["allUsers", "allAuthenticatedUsers"]
        if conf.get("binding") and isinstance(conf.get("binding"), list):
            bindings = conf.get("binding")
            for binding in bindings:
                if binding.get("members") and isinstance(binding.get("members"), list):
                    members = binding.get("members")[0]
                    for member in members:
                        if member in fails:
                            self.evaluated_keys = ["bindings/[0]/members"]
                            return CheckResult.FAILED
                    return CheckResult.PASSED
        return CheckResult.UNKNOWN


check = GooglePolicyIsPrivate()



from __future__ import annotations

import logging
from copy import deepcopy
from pathlib import Path
from typing import Any

from checkov.common.graph.graph_builder import CustomAttributes
from checkov.common.graph.graph_builder.consts import GraphSource, SELF_REFERENCE
from checkov.common.graph.graph_builder.graph_components.block_types import BlockType
from checkov.common.graph.graph_builder.graph_components.blocks import Block
from checkov.common.runners.graph_builder.local_graph import ObjectLocalGraph
from checkov.common.util.consts import START_LINE, END_LINE
from checkov.ansible.graph_builder.graph_components.resource_types import ResourceType
from checkov.ansible.utils import get_scannable_file_paths, TASK_RESERVED_KEYWORDS, parse_file


class AnsibleLocalGraph(ObjectLocalGraph):
    def __init__(self, definitions: dict[str | Path, dict[str, Any] | list[dict[str, Any]]]) -> None:
        super().__init__(definitions=definitions)

        self.source = GraphSource.ANSIBLE

    def _create_vertices(self) -> None:
        for file_path, definition in self.definitions.items():
            if not isinstance(definition, list):
                logging.debug(f"definition of file {file_path} has the wrong type {type(definition)}")
                continue

            file_path = str(file_path)

            for code_block in definition:
                if ResourceType.TASKS in code_block:
                    for task in code_block[ResourceType.TASKS]:
                        self._process_blocks(file_path=file_path, task=task)
                else:
                    self._process_blocks(file_path=file_path, task=code_block)

    def _process_blocks(self, file_path: str, task: Any, prefix: str = "") -> None:
        """Checks for possible block usage"""

        if not task or not isinstance(task, dict):
            return

        if ResourceType.BLOCK in task and isinstance(task[ResourceType.BLOCK], list):
            prefix += f"{ResourceType.BLOCK}."  # with each nested level an extra block prefix is added
            self._create_block_vertices(file_path=file_path, block=task, prefix=prefix)

            for block_task in task[ResourceType.BLOCK]:
                self._process_blocks(file_path=file_path, task=block_task, prefix=prefix)
        else:
            self._create_tasks_vertices(file_path=file_path, task=task, prefix=prefix)

    def _create_tasks_vertices(self, file_path: str, task: Any, prefix: str = "") -> None:
        """Creates tasks vertices"""

        if not task or not isinstance(task, dict):
            return

        # grab the task name at the beginning before trying to find the actual module name
        task_name = task.get("name") or "unknown"

        for name, config in task.items():
            if name in TASK_RESERVED_KEYWORDS:
                continue
            if name in (START_LINE, END_LINE):
                continue
            if isinstance(config, list):
                # either it is actually not an Ansible file or a playbook without tasks refs
                continue

            resource_type = f"{ResourceType.TASKS}.{name}"

            if isinstance(config, str):
                # this happens when modules have no parameters and are directly used with the user input
                # ex. ansible.builtin.command: cat /etc/passwd
                config = {SELF_REFERENCE: config}
            elif config is None:
                # this happens when modules have no parameters and are passed no value
                # ex. amazon.aws.ec2_instance_info:
                config = {
                    START_LINE: task[START_LINE],
                    END_LINE: task[END_LINE],
                }

            attributes = deepcopy(config)
            attributes[CustomAttributes.RESOURCE_TYPE] = resource_type

            # only the module code is relevant for validation,
            # but in the check result the whole task should be visible
            attributes[START_LINE] = task[START_LINE]
            attributes[END_LINE] = task[END_LINE]

            self.vertices.append(
                Block(
                    name=f"{resource_type}.{task_name}",
                    config=config,
                    path=file_path,
                    block_type=BlockType.RESOURCE,
                    attributes=attributes,
                    id=f"{resource_type}.{prefix}{task_name}",
                    source=self.source,
                )
            )

            # no need to further check
            break

    def _create_block_vertices(self, file_path: str, block: dict[str, Any], prefix: str = "") -> None:
        """Creates block vertices"""

        # grab the block name, if it exists
        block_name = block.get("name") or "unknown"

        config = block
        attributes = deepcopy(config)
        attributes[CustomAttributes.RESOURCE_TYPE] = ResourceType.BLOCK
        del attributes[ResourceType.BLOCK]  # the real block content are tasks, which have their own vertices

        self.vertices.append(
            Block(
                name=f"{ResourceType.BLOCK}.{block_name}",
                config=config,
                path=file_path,
                block_type=BlockType.RESOURCE,
                attributes=attributes,
                id=f"{prefix}{block_name}",
                source=self.source,
            )
        )

    def _create_edges(self) -> None:
        return None

    @staticmethod
    def get_files_definitions(root_folder: str | Path) -> dict[str | Path, dict[str, Any] | list[dict[str, Any]]]:
        definitions: "dict[str | Path, dict[str, Any] | list[dict[str, Any]]]" = {}
        file_paths = get_scannable_file_paths(root_folder=root_folder)

        for file_path in file_paths:
            result = parse_file(f=file_path)
            if result is not None:
                definitions[file_path] = result[0]

        return definitions



from __future__ import annotations

from typing import Any

from checkov.common.models.enums import CheckResult, CheckCategories
from checkov.terraform.checks.resource.base_resource_check import BaseResourceCheck


class GCPCloudRunPrivateService(BaseResourceCheck):
    def __init__(self) -> None:
        name = "Ensure that GCP Cloud Run services are not anonymously or publicly accessible"
        id = "CKV_GCP_102"
        supported_resources = ("google_cloud_run_service_iam_member", "google_cloud_run_service_iam_binding")
        categories = (CheckCategories.GENERAL_SECURITY,)
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf: dict[str, list[Any]]) -> CheckResult:
        public_principals = (
            "allUsers",
            "allAuthenticatedUsers"
        )
        # Depending on the terraform resource type -
        # The member config is either a list or single principal
        if self.entity_type == "google_cloud_run_service_iam_member":
            # conf.get returns as a list
            # so we create a string for comparison
            if "member" in conf.keys():
                member = conf.get("member")[0]
                if member in public_principals:
                    return CheckResult.FAILED
                else:
                    return CheckResult.PASSED
        # iam_binding returns a list of principals
        elif self.entity_type == "google_cloud_run_service_iam_binding":
            # Since conf.get returns a list and iam_binding returns a list (nested list)
            # we pull out the members list using the index 0
            if "members" in conf.keys():
                members_list = conf.get("members")[0]
                if any(member in public_principals for member in members_list):
                    return CheckResult.FAILED
                else:
                    return CheckResult.PASSED


check = GCPCloudRunPrivateService()



from abc import ABC
from typing import Dict, List, Any

from cloudsplaining.scan.policy_document import PolicyDocument

from checkov.common.models.enums import CheckResult, CheckCategories
from checkov.terraform.checks.resource.base_resource_check import BaseResourceCheck
from checkov.terraform.checks.utils.base_cloudsplaining_iam_scanner import BaseTerraformCloudsplainingIAMScanner


class BaseTerraformCloudsplainingResourceIAMCheck(BaseResourceCheck, BaseTerraformCloudsplainingIAMScanner, ABC):
    def __init__(self, name: str, id: str) -> None:
        super().__init__(name=name, id=id, categories=[CheckCategories.IAM], supported_resources=["aws_iam_policy"])

    def scan_resource_conf(self, conf: Dict[str, List[Any]]) -> CheckResult:
        return self.scan_conf(conf)

    @property
    def cache_key(self) -> str:
        return self.entity_path

    def should_scan_conf(self, conf: Dict[str, List[Any]]) -> bool:
        return "policy" in conf.keys()

    def convert_to_iam_policy(self, conf: Dict[str, Any]) -> PolicyDocument:
        policy = conf['policy'][0]  # type: Dict[str, Any]
        return PolicyDocument(policy)



from __future__ import annotations

import os
import uuid
from abc import abstractmethod


class BaseGetter:
    def __init__(self, url: str) -> None:
        self.url = url
        self.temp_dir = ""

    def get(self) -> str:
        current_directory = os.getcwd()
        final_directory = os.path.join(current_directory, str((uuid.uuid4()))[:8] + "_checks")
        if not os.path.exists(final_directory):
            os.makedirs(final_directory)
        self.temp_dir = final_directory
        return self.do_get()

    @abstractmethod
    def do_get(self) -> str:
        raise NotImplementedError()



from __future__ import annotations

from typing import Any

from checkov.terraform.checks.resource.base_resource_check import BaseResourceCheck
from checkov.common.models.enums import CheckResult, CheckCategories


class NetworkIPsecAuthAlgorithms(BaseResourceCheck):
    def __init__(self) -> None:
        name = "Ensure IPsec profiles do not specify use of insecure authentication algorithms"
        id = "CKV_PAN_12"
        supported_resources = ('panos_ipsec_crypto_profile', 'panos_panorama_ipsec_crypto_profile')
        categories = (CheckCategories.NETWORKING,)
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf: dict[str, list[Any]]) -> CheckResult:
        # Check there are authentications defined in the resource
        if 'authentications' in conf:

            # Report the area of evaluation
            self.evaluated_keys = ['authentications']

            # Get all the algorithms
            algorithms = conf['authentications']

            # Iterate over each algorithm, as multiple can be defined in "authentications"
            for algo in algorithms:

                # Check for insecure algorithms, including null as a string (not a null value)
                if algo[0] in ('none', 'md5', 'sha1'):

                    # Fail if any insecure algorithms are defined for use
                    return CheckResult.FAILED

            # If no fails have been found, this is a pass
            return CheckResult.PASSED

        # If the mandatory "authentications" attribute is not defined, this is not valid, and will fail during Terraform plan stage, and should therefore be a fail
        return CheckResult.FAILED


check = NetworkIPsecAuthAlgorithms()



from __future__ import annotations

from typing import List, Optional, Any

from checkov.common.output.common import SCADetails

UNFIXABLE_VERSION = "N/A"
OPEN_STATUS = "open"

TWISTCLI_TO_CHECKOV_LANG_NORMALIZATION = {
    "gem": "ruby",
    "nuget": "dotNet"
}


def get_file_path_for_record(rootless_file_path: str) -> str:
    return f"/{rootless_file_path}"


def get_resource_for_record(rootless_file_path: str, package_name: str) -> str:
    return f"{rootless_file_path}.{package_name}"


def get_package_alias(package_name: str, package_version: str) -> str:
    return f"{package_name}@{package_version}"


def get_package_type(package_name: str, package_version: str, sca_details: SCADetails | None = None) -> str:
    if sca_details:
        return str(sca_details.package_types.get(f"{package_name}@{package_version}", ""))
    else:
        return ""


def get_registry_url(package: dict[str, Any]) -> str:
    if "registry" in package:
        return package.get("registry", "")
    return package.get("registryUrl", "")


def normalize_twistcli_language(language: str) -> str:
    """
    part of the language names that are returned by twistcli may be a little differ from those we use in checkov.
    this function's goal is to normalize them
    """
    return TWISTCLI_TO_CHECKOV_LANG_NORMALIZATION.get(language, language)


def should_run_scan(runner_filter_checks: Optional[List[str]]) -> bool:
    return not (runner_filter_checks and all(not (check.startswith("CKV_CVE") or check.startswith("BC_CVE") or check.startswith("BC_LIC")) for check in runner_filter_checks))


def get_fix_version(vulnerability_details: dict[str, Any]) -> str:
    if "fix_version" in vulnerability_details:
        return str(vulnerability_details["fix_version"])

    if "lowest_fixed_version" in vulnerability_details:
        return str(vulnerability_details["lowest_fixed_version"])

    return UNFIXABLE_VERSION



from __future__ import annotations

import os
from typing import Union, Dict, Any, List, Optional, Set
import dpath
import re

from checkov.common.runners.base_runner import strtobool
from checkov.terraform import TFModule
from checkov.terraform.graph_builder.utils import INTERPOLATION_EXPR
from checkov.common.graph.graph_builder.graph_components.blocks import Block
from checkov.common.util.consts import RESOLVED_MODULE_ENTRY_NAME
from checkov.terraform.graph_builder.graph_components.block_types import BlockType
from checkov.terraform.graph_builder.utils import remove_module_dependency_in_path


class TerraformBlock(Block):
    __slots__ = (
        "module_connections",
        "module_dependency",
        "module_dependency_num",
        "source_module",
        "has_dynamic_block",
        "dynamic_attributes",
        "foreach_attrs",
        "source_module_object",
        "for_each_index"
    )

    def __init__(self, name: str, config: Dict[str, Any], path: str, block_type: BlockType, attributes: Dict[str, Any],
                 id: str = "", source: str = "", has_dynamic_block: bool = False, dynamic_attributes: dict[str, Any] | None = None) -> None:
        """
            :param name: unique name given to the terraform block, for example: 'aws_vpc.example_name'
            :param config: the section in tf_definitions that belong to this block
            :param path: the file location of the block
            :param block_type: BlockType
            :param attributes: dictionary of the block's original attributes in the terraform file
        """
        super(TerraformBlock, self).__init__(name, config, path, str(block_type), attributes, id, source, has_dynamic_block, dynamic_attributes)
        self.module_dependency = ""
        self.module_dependency_num = ""
        if path:
            if strtobool(os.getenv('CHECKOV_ENABLE_NESTED_MODULES', 'True')):
                self.path = path
            else:
                self.path, module_dependency, num = remove_module_dependency_in_path(path)
                self.path = os.path.realpath(self.path)
                if module_dependency:
                    self.module_dependency = module_dependency
                    self.module_dependency_num = num
        if attributes.get(RESOLVED_MODULE_ENTRY_NAME):
            del attributes[RESOLVED_MODULE_ENTRY_NAME]
        self.attributes = attributes
        self.module_connections: Dict[str, List[int]] = {}
        self.source_module: Set[int] = set()
        self.has_dynamic_block = has_dynamic_block
        if strtobool(os.getenv('CHECKOV_NEW_TF_PARSER', 'True')):
            self.source_module_object: Optional[TFModule] = None
            self.for_each_index: Optional[Any] = None

    def add_module_connection(self, attribute_key: str, vertex_id: int) -> None:
        self.module_connections.setdefault(attribute_key, []).append(vertex_id)

    def extract_additional_changed_attributes(self, attribute_key: str) -> List[str]:
        # if the `attribute_key` starts with a `for_each.` we know the attribute can't be a dynamic attribute as it
        # represents the for_each of the block, so we don't need extract dynamic changed attributes
        # Fix: https://github.com/bridgecrewio/checkov/issues/4324
        if self.has_dynamic_block and not attribute_key.startswith('for_each'):
            return self._extract_dynamic_changed_attributes(attribute_key)
        return super().extract_additional_changed_attributes(attribute_key)

    def _extract_dynamic_changed_attributes(self, dynamic_attribute_key: str, nesting_prefix='') -> List[str]:
        dynamic_changed_attributes = []
        dynamic_attribute_key_parts = dynamic_attribute_key.split('.')
        try:
            remainder_key_parts = ['start_extract_dynamic_changed_attributes']  # For 1st iteration
            while remainder_key_parts:
                dynamic_for_each_index = dynamic_attribute_key_parts.index('for_each')
                dynamic_content_key_parts, remainder_key_parts = dynamic_attribute_key_parts[:dynamic_for_each_index],\
                    dynamic_attribute_key_parts[dynamic_for_each_index + 1:]
                dynamic_block_name = dynamic_content_key_parts[-1]
                dynamic_content_path = dynamic_content_key_parts + ['content']
                if dpath.search(self.attributes, dynamic_content_path):
                    dynamic_block_content = dpath.get(self.attributes, dynamic_content_path)
                    for key, value in dynamic_block_content.items():
                        key_path = ".".join(filter(None, [nesting_prefix, dynamic_block_name, key]))
                        self._collect_dynamic_dependent_keys(dynamic_block_name, value, key_path, dynamic_content_path, dynamic_changed_attributes)
                dynamic_attribute_key_parts = remainder_key_parts
            return dynamic_changed_attributes
        except ValueError:
            return dynamic_changed_attributes

    def _collect_dynamic_dependent_keys(self, dynamic_block_name: str, value: Union[str, List], key_path: str,
                                        dynamic_content_path: List[str], dynamic_changed_attributes: List[str]) -> None:
        if isinstance(value, str):
            dynamic_ref = f'{dynamic_block_name}.value'
            interpolation_matches = re.findall(INTERPOLATION_EXPR, value)
            for match in interpolation_matches:
                if dynamic_ref in match:
                    dynamic_changed_attributes.append(key_path)
        elif isinstance(value, list):
            for idx, sub_value in enumerate(value):
                self._collect_dynamic_dependent_keys(
                    dynamic_block_name, sub_value, f'{key_path}.{idx}', dynamic_content_path, dynamic_changed_attributes)
        elif isinstance(value, dict):
            for sub_key, sub_value in value.items():
                if isinstance(sub_value, dict) and 'content' in sub_value.keys() and 'for_each' in sub_value.keys():
                    nested_dynamic_block_key_path = f'{".".join(dynamic_content_path)}.dynamic.{sub_key}.for_each'
                    dynamic_changed_attributes.extend(self._extract_dynamic_changed_attributes(nested_dynamic_block_key_path, nesting_prefix=dynamic_block_name))
                else:
                    self._collect_dynamic_dependent_keys(
                        dynamic_block_name, sub_value, f'{key_path}.{sub_key}', dynamic_content_path, dynamic_changed_attributes)

    def find_attribute(self, attribute: Optional[Union[str, List[str]]]) -> Optional[str]:
        """
        :param attribute: key to search in self.attributes
        The function searches for  attribute in self.attribute. It might not exist if the block is variable or output,
        or its search path might be different if its a resource.
        :return: the actual attribute key or None
        """
        if not attribute:
            return None

        if self.attributes.get(attribute[0]):
            return attribute[0]

        if self.block_type == BlockType.VARIABLE:
            return "default" if self.attributes.get("default") else None

        if self.block_type == BlockType.OUTPUT:
            return "value" if self.attributes.get("value") else None

        if self.block_type == BlockType.RESOURCE and len(attribute) > 1:
            # handle cases where attribute_at_dest == ['aws_s3_bucket.template_bucket', 'acl']
            if self.name == attribute[0] and self.attributes.get(attribute[1]):
                return attribute[1]

        return None

    def update_list_attribute(self, attribute_key: str, attribute_value: Any) -> None:
        """Updates list attributes with their index

        This needs to be overridden, because of our hcl parser adding a list around any value
        """

        if attribute_key not in self.attributes or isinstance(self.attributes[attribute_key][0], list):
            # sometimes the attribute_value is a list and replaces the whole value of the key, which makes it a normal value
            # ex. attribute_value = ["xyz"] and self.attributes[attribute_key][0] = "xyz"
            for idx, value in enumerate(attribute_value):
                self.attributes[f"{attribute_key}.{idx}"] = value

    @classmethod
    def get_inner_attributes(
        cls,
        attribute_key: str,
        attribute_value: Union[str, List[str], Dict[str, Any]],
        strip_list: bool = True
    ) -> Dict[str, Any]:
        if strip_list and isinstance(attribute_value, list) and len(attribute_value) == 1:
            attribute_value = attribute_value[0]

        return super().get_inner_attributes(
            attribute_key=attribute_key,
            attribute_value=attribute_value,
        )

    def to_dict(self):
        return {
            'attributes': self.attributes,
            'block_type': self.block_type,
            'breadcrumbs': self.breadcrumbs,
            'config': self.config,
            'id': self.id,
            'module_connections': self.module_connections,
            'module_dependency': self.module_dependency,
            'module_dependency_num': self.module_dependency_num,
            'name': self.name,
            'path': self.path,
            'source': self.source,
            'source_module': list(self.source_module)
        }



from checkov.common.models.enums import CheckResult, CheckCategories
from checkov.cloudformation.checks.resource.base_resource_check import BaseResourceCheck


class IMDSv1Disabled(BaseResourceCheck):
    def __init__(self):
        name = "Ensure Instance Metadata Service Version 1 is not enabled"
        id = "CKV_AWS_79"
        supported_resources = ['AWS::EC2::LaunchTemplate']
        categories = [CheckCategories.IAM]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        # IMDS can be disabled or IMDSv2 can be enabled
        if 'Properties' in conf.keys():
            properties = conf['Properties']
            if 'LaunchTemplateData' in properties.keys():
                launch_template_data = properties['LaunchTemplateData']
            if 'MetadataOptions' in launch_template_data.keys():
                metadata_options = launch_template_data['MetadataOptions']
                if 'HttpEndpoint' in metadata_options.keys():
                    if metadata_options['HttpEndpoint'] == "disabled":
                        return CheckResult.PASSED
                if 'HttpTokens' in metadata_options.keys():
                    if metadata_options['HttpTokens'] == "required":  # checkov:skip=CKV_SECRET_6 false positive
                        return CheckResult.PASSED
        return CheckResult.FAILED


check = IMDSv1Disabled()



from __future__ import annotations

import concurrent.futures
import logging
import multiprocessing
import os
import platform
from typing import Any, List, Generator, Iterator, Callable, Optional, TypeVar, TYPE_CHECKING

if TYPE_CHECKING:
    from multiprocessing.connection import Connection

_T = TypeVar("_T")


class ParallelRunner:
    def __init__(self, workers_number: int | None = None) -> None:
        self.workers_number = (workers_number if workers_number else os.cpu_count()) or 1
        self.os = platform.system()

    def run_function(self, func: Callable[[Any], _T], items: List[Any], group_size: Optional[int] = None, run_multiprocess: Optional[bool] = False) -> Iterator[_T]:
        if self.os == 'Windows' or (not run_multiprocess and os.getenv("PYCHARM_HOSTED") == "1"):
            # PYCHARM_HOSTED env variable equals 1 when debugging via jetbrains IDE.
            # To prevent JetBrains IDE from crashing on debug use multi threading
            # Override this condition if run_multiprocess is set to True
            return self._run_function_multithreaded(func, items)
        else:
            return self._run_function_multiprocess(func, items, group_size)

    def _run_function_multiprocess(self, func: Callable[[Any], Any], items: List[Any], group_size: Optional[int]) \
            -> Generator[Any, None, None]:
        if not group_size:
            group_size = int(len(items) / self.workers_number) + 1
        groups_of_items = [items[i: i + group_size] for i in range(0, len(items), group_size)]
        logging.info(f"[_run_function_multiprocess] starting with {self.workers_number} workers")

        def func_wrapper(original_func: Callable[[Any], Any], items_group: List[Any], connection: Connection) -> None:
            for item in items_group:
                try:
                    result = original_func(item)
                except Exception:
                    logging.error(
                        f"Failed to invoke function {func.__code__.co_filename.replace('.py', '')}.{func.__name__} with {item}"
                        , exc_info=True
                    )
                    result = None

                connection.send(result)
            connection.close()

        processes = []
        for group_of_items in groups_of_items:
            parent_conn, child_conn = multiprocessing.Pipe(duplex=False)
            process = multiprocessing.get_context("fork").Process(target=func_wrapper,
                                                                  args=(func, group_of_items, child_conn))
            processes.append((process, parent_conn, len(group_of_items)))
            process.start()

        for _, parent_conn, group_len in processes:
            for _ in range(group_len):
                try:
                    yield parent_conn.recv()
                except EOFError:
                    pass

    def _run_function_multithreaded(self, func: Callable[[Any], _T], items: List[Any]) -> Iterator[_T]:
        logging.info(f"[_run_function_multithreaded] starting with {self.workers_number} workers")
        with concurrent.futures.ThreadPoolExecutor(max_workers=self.workers_number) as executor:
            return executor.map(func, items)


parallel_runner = ParallelRunner()



from checkov.terraform.checks.resource.base_resource_value_check import BaseResourceValueCheck
from checkov.common.models.enums import CheckCategories


class ElasticsearchEncryption(BaseResourceValueCheck):
    def __init__(self) -> None:
        name = "Ensure all data stored in the Elasticsearch is securely encrypted at rest"
        id = "CKV_AWS_5"
        supported_resources = ('aws_elasticsearch_domain', 'aws_opensearch_domain')
        categories = (CheckCategories.ENCRYPTION,)
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def get_inspected_key(self) -> str:
        return "encrypt_at_rest/[0]/enabled"


check = ElasticsearchEncryption()



from __future__ import annotations

import logging
import os
from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Any

from dockerfile_parse.constants import COMMENT_INSTRUCTION

from checkov.common.runners.base_runner import filter_ignored_paths
from checkov.common.util.dockerfile import is_docker_file
from checkov.common.util.suppression import collect_suppressions_for_context
from checkov.dockerfile.parser import parse

if TYPE_CHECKING:
    from dockerfile_parse.parser import _Instruction  # only in extra_stubs
    from typing_extensions import Literal

DOCKERFILE_STARTLINE: Literal["startline"] = "startline"
DOCKERFILE_ENDLINE: Literal["endline"] = "endline"
DOCKERFILE_VALUE: Literal["value"] = "value"


def get_scannable_file_paths(
    root_folder: str | Path | None = None, excluded_paths: list[str] | None = None
) -> set[str]:
    """Finds Dockerfiles"""

    file_paths: "set[str]" = set()
    if not root_folder:
        return file_paths

    for root, d_names, f_names in os.walk(root_folder):
        filter_ignored_paths(root, d_names, excluded_paths)
        filter_ignored_paths(root, f_names, excluded_paths)
        for file in f_names:
            if is_docker_file(file):
                file_path = os.path.join(root, file)
                file_paths.add(file_path)

    return file_paths


def get_files_definitions(
    files: Iterable[str], filepath_fn: Callable[[str], str] | None = None
) -> tuple[dict[str, dict[str, list[_Instruction]]], dict[str, list[str]]]:
    """Parses Dockerfiles into its definitions and raw data"""

    definitions = {}
    definitions_raw = {}

    for file in files:
        try:
            result = parse(file)

            path = filepath_fn(file) if filepath_fn else file
            definitions[path], definitions_raw[path] = result
        except TypeError:
            logging.info(f"Dockerfile skipping {file} as it is not a valid dockerfile template")
        except UnicodeDecodeError:
            logging.info(f"Dockerfile skipping {file} as it can't be read as text file")

    return definitions, definitions_raw


def get_abs_path(root_folder: str | None, file_path: str) -> str:
    """Creates the abs path

    There are a few cases here. If -f was used, there could be a leading / because it's an absolute path,
    or there will be no leading slash; root_folder will always be none.
    If -d is used, root_folder will be the value given, and -f will start with a / (hardcoded above).
    The goal here is simply to get a valid path to the file (which docker_file_path does not always give).
    """

    if root_folder and file_path.startswith("/"):
        # remove the leading slash, if it exists
        file_path = file_path[1:]

    path_to_convert = os.path.join(root_folder, file_path) if root_folder else file_path

    return os.path.abspath(path_to_convert)


def build_definitions_context(
    definitions: dict[str, dict[str, list[_Instruction]]],
    definitions_raw: dict[str, list[str]]
) -> dict[str, dict[str, Any]]:
    definitions_context: dict[str, dict[str, Any]] = {}

    for file_path, definition in definitions.items():
        file_path = str(file_path)
        definitions_context[file_path] = {}
        skipped_checks = []
        if COMMENT_INSTRUCTION in definition:
            # collect skipped check comments
            comments = definition[COMMENT_INSTRUCTION]
            comment_lines = [(comment[DOCKERFILE_STARTLINE], comment[DOCKERFILE_VALUE]) for comment in comments]
            skipped_checks = collect_suppressions_for_context(code_lines=comment_lines)

        for instruction_name, instructions in definition.items():
            if instruction_name == COMMENT_INSTRUCTION:
                continue

            definitions_context[file_path][instruction_name] = []
            for instruction in instructions:
                start_line = instruction[DOCKERFILE_STARTLINE]
                end_line = instruction[DOCKERFILE_ENDLINE]
                code_lines = [
                    (line + 1, definitions_raw[file_path][line])
                    for line in range(start_line, end_line + 1)
                ]
                definition_resource = {
                    "start_line": start_line + 1,  # lines start with index 0
                    "end_line": end_line + 1,
                    "code_lines": code_lines,
                    "skipped_checks": skipped_checks,
                }
                definitions_context[file_path][instruction_name].append(definition_resource)

    return definitions_context



from enum import Enum


class SolverType(str, Enum):
    # A solver is a class that resolves YAML syntax into a graph query
    # It can have the following types:
    ATTRIBUTE = "ATTRIBUTE"
    # An attribute query, i.e. id equals 3

    COMPLEX = "COMPLEX"
    # A combination of queries, i.e. <SOME_QUERY> AND <ANOTHER_QUERY>

    CONNECTION = "CONNECTION"
    # A connection between two entities, i.e. ec2 instance connected to security group

    # TODO: merge with COMPLEX
    COMPLEX_CONNECTION = "COMPLEX_CONNECTION"
    # A combination of CONNECTION solver and any other solver

    FILTER = "FILTER"
    # Filters results according to specific value / type, i.e. resource type is aws_s3_bucket


class Operators:
    ANY = 'any'
    EXISTS = 'exists'
    ONE_EXISTS = 'one_exists'
    NOT_EXISTS = 'not_exists'
    CONTAINS = 'contains'
    NOT_CONTAINS = 'not_contains'
    ENDING_WITH = 'ending_with'
    NOT_ENDING_WITH = 'not_ending_with'
    EQUALS = 'equals'
    NOT_EQUALS = 'not_equals'
    REGEX_MATCH = 'regex_match'
    NOT_REGEX_MATCH = 'not_regex_match'
    GREATER_THAN = 'greater_than'
    GREATER_THAN_OR_EQUAL = 'greater_than_or_equal'
    LESS_THAN = 'less_than'
    LESS_THAN_OR_EQUAL = 'less_than_or_equal'
    STARTING_WITH = 'starting_with'
    NOT_STARTING_WITH = 'not_starting_with'
    SUBSET = 'subset'
    NOT_SUBSET = 'not_subset'
    WITHIN = 'within'
    NOT_WITHIN = 'not_within'
    AND = 'and'
    OR = 'or'
    NOT = 'not'
    JSONPATH_EQUALS = 'jsonpath_equals'
    JSONPATH_NOT_EQUALS = 'jsonpath_not_equals'
    JSONPATH_EXISTS = 'jsonpath_exists'
    JSONPATH_NOT_EXISTS = 'jsonpath_not_exists'
    IS_EMPTY = 'is_empty'
    IS_NOT_EMPTY = 'is_not_empty'
    LENGTH_EQUALS = 'length_equals'
    LENGTH_NOT_EQUALS = 'length_not_equals'
    LENGTH_GREATER_THAN = 'length_greater_than'
    LENGTH_GREATER_THAN_OR_EQUAL = 'length_greater_than_or_equal'
    LENGTH_LESS_THAN = 'length_less_than'
    LENGTH_LESS_THAN_OR_EQUAL = 'length_less_than_or_equal'
    IS_TRUE = 'is_true'
    IS_FALSE = 'is_false'
    INTERSECTS = 'intersects'
    NOT_INTERSECTS = 'not_intersects'
    EQUALS_IGNORE_CASE = 'equals_ignore_case'
    NOT_EQUALS_IGNORE_CASE = 'not_equals_ignore_case'
    RANGE_INCLUDES = 'range_includes'
    RANGE_NOT_INCLUDES = 'range_not_includes'
    NUMBER_OF_WORDS_EQUALS = 'number_of_words_equals'
    NUMBER_OF_WORDS_NOT_EQUALS = 'number_of_words_not_equals'
    NUMBER_OF_WORDS_GREATER_THAN = 'number_of_words_greater_than'
    NUMBER_OF_WORDS_GREATER_THAN_OR_EQUAL = 'number_of_words_greater_than_or_equal'
    NUMBER_OF_WORDS_LESS_THAN = 'number_of_words_less_than'
    NUMBER_OF_WORDS_LESS_THAN_OR_EQUAL = 'number_of_words_less_than_or_equal'



from __future__ import annotations

import logging
from pathlib import Path
from typing import Any

from yaml import YAMLError

from checkov.kubernetes.parser import k8_yaml, k8_json
from checkov.kubernetes.parser.validatior import K8sValidator

logger = logging.getLogger(__name__)


def parse(filename: str) -> tuple[list[dict[str, Any]], list[tuple[int, str]]] | None:
    template = None
    template_lines: "list[tuple[int, str]]" = []
    valid_templates = []
    try:
        if filename.endswith((".yaml", ".yml")):
            (template, template_lines) = k8_yaml.load(Path(filename))
        if filename.endswith(".json"):
            (template, template_lines) = k8_json.load(Path(filename))
        if template:
            if isinstance(template, list):
                for i, t in enumerate(template):
                    is_valid, reason = K8sValidator.is_valid_template(t)
                    if is_valid:
                        valid_templates.append(t)
                    else:
                        logging.debug(f"template {i} from file {filename} is not a valid k8s template, reason: {reason}")
            else:
                return None
        else:
            return None
    except IOError as e:
        if e.errno == 2:
            logger.error('Template file not found: %s', filename)
            return None
        elif e.errno == 21:
            logger.error('Template references a directory, not a file: %s',
                         filename)
            return None
        elif e.errno == 13:
            logger.error('Permission denied when accessing template file: %s',
                         filename)
            return None
    except UnicodeDecodeError:
        logger.error('Cannot read file contents: %s', filename)
        return None
    except YAMLError:
        if filename.endswith(".yaml") or filename.endswith(".yml"):
            logger.debug('Cannot read file contents: %s - is it a yaml?', filename)
        return None

    return valid_templates, template_lines



from __future__ import annotations

import os
import logging
from http import HTTPStatus
from typing import List, Dict, TYPE_CHECKING

import requests
from requests.exceptions import HTTPError
from urllib.parse import urlparse

from checkov.common.models.consts import TFC_HOST_NAME
from checkov.common.goget.registry.get_registry import RegistryGetter
from checkov.common.util.http_utils import DEFAULT_TIMEOUT
from checkov.terraform.module_loading.content import ModuleContent
from checkov.terraform.module_loading.loader import ModuleLoader
from checkov.terraform.module_loading.loaders.versions_parser import (
    order_versions_in_descending_order,
    get_version_constraints
)

if TYPE_CHECKING:
    from checkov.terraform.module_loading.module_params import ModuleParams

# https://developer.hashicorp.com/terraform/language/modules/sources#fetching-archives-over-http
MODULE_ARCHIVE_EXTENSIONS = ["zip", "tar.bz2", "tar.gz", "tgz", "tar.xz", "txz"]


class RegistryLoader(ModuleLoader):
    modules_versions_cache: Dict[str, List[str]] = {}  # noqa: CCE003  # public data

    def __init__(self) -> None:
        super().__init__()

    def discover(self, module_params: ModuleParams) -> None:
        module_params.tf_host_name = os.getenv("TF_HOST_NAME", TFC_HOST_NAME)
        module_params.token = os.getenv("TF_REGISTRY_TOKEN", "")
        tfc_token = os.getenv("TFC_TOKEN")
        if tfc_token:
            self.logger.warn("Environment variable TFC_TOKEN will be deprecated in the future. Please use TF_REGISTRY_TOKEN instead.")
            module_params.token = tfc_token

    def _is_matching_loader(self, module_params: ModuleParams) -> bool:
        # https://developer.hashicorp.com/terraform/language/modules/sources#github
        if module_params.module_source.startswith(("github.com", "bitbucket.org", "git::", "git@github.com")):
            return False
        self._process_inner_registry_module(module_params)
        # determine tf api endpoints
        self._determine_tf_api_endpoints(module_params)
        # If versions for a module are cached, determine the best version and return True.
        # If versions are not cached, get versions, then determine the best version and return True.
        # Best version needs to be determined here for setting most accurate dest_dir.
        if module_params.tf_modules_versions_endpoint in RegistryLoader.modules_versions_cache.keys():
            module_params.best_version = self._find_best_version(module_params)
            return True
        if not self._cache_available_versions(module_params):
            return False
        module_params.best_version = self._find_best_version(module_params)
        if not module_params.inner_module and module_params.tf_host_name:
            module_params.dest_dir = os.path.join(module_params.root_dir, module_params.external_modules_folder_name,
                                                  module_params.tf_host_name, *module_params.module_source.split("/"),
                                                  module_params.best_version)
        if os.path.exists(module_params.dest_dir):
            return True
        # verify cache again after refresh
        if module_params.tf_modules_versions_endpoint in RegistryLoader.modules_versions_cache.keys():
            return True
        return False

    def _load_module(self, module_params: ModuleParams) -> ModuleContent:
        if module_params.best_version:
            best_version = module_params.best_version
        else:
            if self._cache_available_versions(module_params):
                module_params.best_version = self._find_best_version(module_params)
        if os.path.exists(module_params.dest_dir):
            return ModuleContent(dir=module_params.dest_dir)
        elif not module_params.tf_modules_endpoint:
            return ModuleContent(dir=None)

        request_download_url = "/".join((module_params.tf_modules_endpoint, module_params.module_source, best_version, "download"))
        logging.debug(f"Best version for {module_params.module_source} is {best_version} based on the version constraint {module_params.version}.")
        logging.debug(f"Module download url: {request_download_url}")
        try:
            response = requests.get(
                url=request_download_url,
                headers={"Authorization": f"Bearer {module_params.token}"} if module_params.token else None,
                timeout=DEFAULT_TIMEOUT
            )
            response.raise_for_status()
        except HTTPError as e:
            self.logger.warning(e)
            if response.status_code != HTTPStatus.OK and response.status_code != HTTPStatus.NO_CONTENT:
                return ModuleContent(dir=None)
        # https://www.terraform.io/registry/api-docs#download-source-code-for-a-specific-module-version
        module_download_url = response.headers.get('X-Terraform-Get', '')
        self.logger.debug(f"X-Terraform-Get: {module_download_url}")
        module_download_url = self._normalize_module_download_url(module_params, module_download_url)
        self.logger.debug(f"Cloning module from normalized url {module_download_url}")
        if self._is_download_url_archive(module_download_url):
            try:
                registry_getter = RegistryGetter(module_download_url)
                registry_getter.temp_dir = module_params.dest_dir
                registry_getter.do_get()
                return_dir = module_params.dest_dir
            except Exception as e:
                str_e = str(e)
                if 'File exists' not in str_e and 'already exists and is not an empty directory' not in str_e:
                    self.logger.error(f"failed to get {module_params.module_source} because of {e}")
                    return ModuleContent(dir=None, failed_url=module_params.module_source)
            if module_params.inner_module:
                return_dir = os.path.join(module_params.dest_dir, module_params.inner_module)
            return ModuleContent(dir=return_dir)
        else:
            return ModuleContent(dir=None, next_url=response.headers.get("X-Terraform-Get", ""))

    def _find_module_path(self, module_params: ModuleParams) -> str:
        # to determine the exact path here would be almost a duplicate of the git_loader functionality
        return ""

    def _find_best_version(self, module_params: ModuleParams) -> str:
        versions_by_size = RegistryLoader.modules_versions_cache.get(module_params.tf_modules_versions_endpoint, [])  # type:ignore[arg-type]  # argument can be None
        if module_params.version == "latest":
            module_params.version = versions_by_size[0]
        elif module_params.version is None:
            return "latest"

        version_constraints = get_version_constraints(module_params.version)
        num_of_matches = 0
        for version in versions_by_size:
            for version_constraint in version_constraints:
                if not version_constraint.versions_matching(version):
                    break
                else:
                    num_of_matches += 1
            if num_of_matches == len(version_constraints):
                return version
            else:
                num_of_matches = 0
        return "latest"

    def _cache_available_versions(self, module_params: ModuleParams) -> bool:
        # Get all available versions for a module in the registry and cache them.
        # Returns False on failure.
        if not module_params.tf_modules_versions_endpoint:
            return False

        try:
            response = requests.get(
                url=module_params.tf_modules_versions_endpoint,
                headers={"Authorization": f"Bearer {module_params.token}"} if module_params.token else None,
                timeout=DEFAULT_TIMEOUT,
            )
            response.raise_for_status()
            available_versions = [
                v.get("version") for v in response.json().get("modules", [{}])[0].get("versions", {})
            ]
            RegistryLoader.modules_versions_cache[module_params.tf_modules_versions_endpoint] = order_versions_in_descending_order(
                available_versions)
            return True
        except HTTPError as e:
            self.logger.debug(e)
            return False

    def _process_inner_registry_module(self, module_params: ModuleParams) -> None:
        # Check if the source has '//' in it. If it does, it indicates a reference for an inner module.
        # Example: "terraform-aws-modules/security-group/aws//modules/http-80" =>
        #    module_source = terraform-aws-modules/security-group/aws
        #    dest_dir = modules/http-80
        module_source_components = module_params.module_source.split("//")
        if len(module_source_components) > 1:
            module_params.module_source = module_source_components[0]
            module_params.dest_dir = module_params.dest_dir.split("//")[0]
            module_params.inner_module = module_source_components[1]

    def _determine_tf_api_endpoints(self, module_params: ModuleParams) -> None:
        """
        Determines terraform registry endpoints - tf_host_name, tf_modules_endpoint, tf_modules_versions_endpoint
        """
        if module_params.tf_host_name and module_params.module_source.startswith(module_params.tf_host_name):
            # check if module source supports native Terraform services
            # https://www.terraform.io/internals/remote-service-discovery#remote-service-discovery
            module_params.module_source = module_params.module_source.replace(f"{module_params.tf_host_name}/", "")
            try:
                response = requests.get(
                    url=f"https://{module_params.tf_host_name}/.well-known/terraform.json",
                    timeout=DEFAULT_TIMEOUT,
                )
                response.raise_for_status()
            except HTTPError as e:
                self.logger.debug(e)
                if response.status_code != HTTPStatus.OK and response.status_code != HTTPStatus.NO_CONTENT:
                    return None

            self.logger.debug(f"Service discovery response: {response.json()}")
            module_params.tf_modules_endpoint = f"https://{module_params.tf_host_name}{response.json().get('modules.v1')}"
        else:
            # use terraform cloud host name and url for the public registry
            module_params.tf_host_name = TFC_HOST_NAME
            module_params.tf_modules_endpoint = "https://registry.terraform.io/v1/modules"
        module_params.tf_modules_versions_endpoint = "/".join((module_params.tf_modules_endpoint, module_params.module_source, "versions"))

    def _normalize_module_download_url(self, module_params: ModuleParams, module_download_url: str) -> str:
        if not urlparse(module_download_url).netloc:
            module_download_url = f"https://{module_params.tf_host_name}{module_download_url}"
        return module_download_url

    @staticmethod
    def _is_download_url_archive(module_download_url: str) -> bool:
        for extension in MODULE_ARCHIVE_EXTENSIONS:
            if module_download_url.endswith(extension):
                return True
        query_params_str = urlparse(module_download_url).query
        if query_params_str:
            query_params = query_params_str.split("&")
            for query_param in query_params:
                if query_param.startswith("archive="):
                    return True
        return False


loader = RegistryLoader()



from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.terraform.checks.resource.base_resource_check import BaseResourceCheck


class AMILaunchIsShared(BaseResourceCheck):
    def __init__(self):
        name = "Ensure to Limit AMI launch Permissions"
        id = "CKV_AWS_205"
        supported_resources = ['aws_ami_launch_permission']
        categories = [CheckCategories.GENERAL_SECURITY]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        return CheckResult.FAILED


check = AMILaunchIsShared()



from __future__ import annotations

from typing import Any

from checkov.common.models.enums import CheckResult, CheckCategories
from checkov.arm.base_resource_check import BaseResourceCheck


class AzureManagedDiscEncryption(BaseResourceCheck):
    def __init__(self) -> None:
        name = "Ensure Azure managed disk have encryption enabled"
        id = "CKV_AZURE_2"
        supported_resources = ("Microsoft.Compute/disks",)
        categories = (CheckCategories.ENCRYPTION,)
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf: dict[str, Any]) -> CheckResult:
        if "properties" in conf:
            if "encryptionSettingsCollection" in conf["properties"]:
                if "enabled" in conf["properties"]["encryptionSettingsCollection"]:
                    if str(conf["properties"]["encryptionSettingsCollection"]["enabled"]).lower() == "true":
                        return CheckResult.PASSED
            elif "encryptionSettings" in conf["properties"]:
                if "enabled" in conf["properties"]["encryptionSettings"]:
                    if str(conf["properties"]["encryptionSettings"]["enabled"]).lower() == "true":
                        return CheckResult.PASSED
        return CheckResult.FAILED


check = AzureManagedDiscEncryption()



from __future__ import annotations

from typing import Any

from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.kubernetes.checks.resource.base_spec_check import BaseK8Check


class AllowedCapabilities(BaseK8Check):
    def __init__(self) -> None:
        # CIS-1.5 5.2.8
        name = "Do not allow containers with added capability"
        # No capabilities may be added beyond the default set
        # https://kubernetes.io/docs/concepts/policy/pod-security-policy/#capabilities
        # Location: PodSecurityPolicy.spec.allowedCapabilities
        id = "CKV_K8S_24"
        supported_kind = ("PodSecurityPolicy",)
        categories = (CheckCategories.KUBERNETES,)
        super().__init__(name=name, id=id, categories=categories, supported_entities=supported_kind)

    def scan_spec_conf(self, conf: dict[str, Any]) -> CheckResult:
        spec = conf.get("spec")
        if spec and spec.get("allowedCapabilities"):
            return CheckResult.FAILED
        return CheckResult.PASSED


check = AllowedCapabilities()



from checkov.common.models.enums import CheckCategories
from checkov.terraform.checks.resource.gcp.AbsGoogleIAMMemberDefaultServiceAccount import AbsGoogleIAMMemberDefaultServiceAccount


class GoogleFolderMemberDefaultServiceAccount(AbsGoogleIAMMemberDefaultServiceAccount):
    def __init__(self) -> None:
        name = "Ensure Default Service account is not used at a folder level"
        id = "CKV_GCP_48"
        supported_resources = ('google_folder_iam_member', 'google_folder_iam_binding')
        categories = (CheckCategories.IAM,)
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)


check = GoogleFolderMemberDefaultServiceAccount()



from __future__ import annotations

import logging
from pathlib import Path
from typing import Any

from charset_normalizer import from_path
from yaml.scanner import ScannerError
from yaml import YAMLError

from checkov.common.parsers.json import parse as json_parse
from checkov.common.parsers.yaml import loader


LOGGER = logging.getLogger(__name__)


def parse(filename: str) -> tuple[dict[str, Any], list[tuple[int, str]]] | tuple[None, None]:
    """Decode filename into an object"""

    template = None
    template_lines = None
    try:
        template, template_lines = load(filename)
    except IOError as e:
        if e.errno == 2:
            LOGGER.error(f"Template file not found: {filename}")
        elif e.errno == 21:
            LOGGER.error(f"Template references a directory, not a file: {filename}")
        elif e.errno == 13:
            LOGGER.error(f"Permission denied when accessing template file: {filename}")
    except UnicodeDecodeError:
        LOGGER.error(f"Cannot read file contents: {filename}")
    except ScannerError as err:
        if err.problem in ("found character '\\t' that cannot start any token", "found unknown escape character"):
            try:
                result = json_parse(filename, allow_nulls=False)
                if result:
                    template, template_lines = result  # type:ignore[assignment]  # this is handled by the next line
                    if isinstance(template, list):
                        # should not happen and is more relevant for type safety
                        template = template[0]
            except Exception:
                LOGGER.error(f"Template {filename} is malformed: {err.problem}")
                LOGGER.error(f"Tried to parse {filename} as JSON", exc_info=True)
    except YAMLError:
        pass

    if template is None or template_lines is None:
        return None, None

    return template, template_lines


def load(filename: Path | str) -> tuple[dict[str, Any], list[tuple[int, str]]]:
    """
    Load the given JSON/YAML file
    """

    file_path = filename if isinstance(filename, Path) else Path(filename)

    try:
        content = file_path.read_text()
    except UnicodeDecodeError:
        logging.debug(f"Encoding for file {file_path} is not UTF-8, trying to detect it")
        content = str(from_path(file_path).best())

    if not all(key in content for key in ("$schema", "contentVersion")):
        return {}, []

    file_lines = [(idx + 1, line) for idx, line in enumerate(content.splitlines(keepends=True))]

    template: "dict[str, Any] | list[dict[str, Any]]" = loader.loads(content=content)
    if not template:
        template = {}
    if isinstance(template, list):
        template = template[0]

    return template, file_lines



from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Tuple, Optional, List

import jmespath
import logging
import re

import yaml
from yaml import YAMLError

from checkov.cloudformation.parser import cfn_yaml
from checkov.cloudformation.context_parser import ContextParser
from checkov.cloudformation.parser.cfn_yaml import CfnParseError
from checkov.common.models.consts import SLS_DEFAULT_VAR_PATTERN
from checkov.common.parsers.node import DictNode, StrNode

logger = logging.getLogger(__name__)

IAM_ROLE_STATEMENTS_TOKEN = 'iamRoleStatements'  # nosec  # checkov:skip=CKV_SECRET_6 false positive
CFN_RESOURCES_TOKEN = 'resources'  # nosec
PROVIDER_TOKEN = 'provider'  # nosec
FUNCTIONS_TOKEN = 'functions'  # nosec
ENVIRONMENT_TOKEN = 'environment'  # nosec
STACK_TAGS_TOKEN = 'stackTags'  # nosec
TAGS_TOKEN = 'tags'  # nosec
SUPPORTED_PROVIDERS = ['aws']

QUOTED_WORD_SYNTAX = re.compile(r"(?:('|\").*?\1)")
FILE_LOCATION_PATTERN = re.compile(r'^file\(([^?%*:|"<>]+?)\)')


def parse(filename):
    template = None
    template_lines = None
    try:
        (template, template_lines) = cfn_yaml.load(filename, cfn_yaml.ContentType.SLS)
        if not template or not is_checked_sls_template(template):
            return
    except IOError as e:
        if e.errno == 2:
            logger.error('Template file not found: %s', filename)
            return
        elif e.errno == 21:
            logger.error('Template references a directory, not a file: %s',
                         filename)
            return
        elif e.errno == 13:
            logger.error('Permission denied when accessing template file: %s',
                         filename)
            return
    except UnicodeDecodeError:
        logger.error('Cannot read file contents: %s', filename)
        return
    except CfnParseError:
        logger.warning(f"Failed to parse file {filename} because it isn't a valid template")
        return
    except YAMLError:
        logger.warning(f"Failed to parse file {filename} as a yaml")
        return

    process_variables(template, filename)

    return template, template_lines


def is_checked_sls_template(template):
    if template.__contains__('provider'):
        # Case provider is a dictionary
        if isinstance(template['provider'], DictNode):
            if template['provider'].get('name', '').lower() not in SUPPORTED_PROVIDERS:
                return False
        # Case provider is direct provider name
        if isinstance(template['provider'], StrNode):
            if template['provider'] not in SUPPORTED_PROVIDERS:
                return False
        return True
    return False


def template_contains_cfn_resources(template):
    if template.__contains__(CFN_RESOURCES_TOKEN) and isinstance(template[CFN_RESOURCES_TOKEN], DictNode):
        if template[CFN_RESOURCES_TOKEN].get('Resources'):
            return True
    return False


def template_contains_key(template, key):
    if ContextParser.search_deep_keys(key, template, []):
        return True
    return False


def process_variables(template, filename):
    """
Modifies the template data in-place to resolve variables.
    """

    file_data_cache = {}
    service_file_directory = os.path.dirname(filename)

    var_pattern = jmespath.search("provider.variableSyntax", template)
    if var_pattern is not None:
        # Remove to prevent self-matching during processing
        del template["provider"]["variableSyntax"]
    else:
        var_pattern = SLS_DEFAULT_VAR_PATTERN
    compiled_var_pattern = re.compile(var_pattern)

    # Processing is done in a loop to deal with chained references and the like.
    # Loop while the data is being changed, stop when no more changes are happening.
    # To ensure there's not some kind of oscillation, a cap of 25 passes is in place.
    # More than a couple loops isn't normally expected.
    # NOTE: If this approach proves to be a performance liability, a DAG will be needed.
    loop_count = 0
    for _ in range(0, 25):
        loop_count += 1
        made_change = False

        if process_variables_loop(template, compiled_var_pattern,
                                  # vt = var type
                                  # vl = var location
                                  # ft = fallback var type
                                  # fl = fallback var location
                                  lambda vt, vl, ft, fl: _load_var_data(vt, vl, ft, fl,
                                                                        file_data_cache,
                                                                        template, service_file_directory)):
            made_change = True

        if not made_change:
            break
    logger.debug("Processing of %s variables took %d loop iterations", filename, loop_count)

    return template


def process_variables_loop(template, var_pattern, param_lookup_function):
    """
Generic processing loop for variables.
    :param template:                The dictionary currently being processed. This function will
                                    be called recursively starting at dict provided.
    :param var_pattern:             A compiled regex pattern which should name match groups
                                    if they are needed for looking up the data source.
    :param param_lookup_function:   A Callable taking four arguments:
                                    1) the type (e.g., "self", "file(/path/to/file.yml)")
                                    2) the location (e.g., "custom.my_property")
                                    3) fallback var type (same as above plus None for static value)
                                    4) fallback var location or value if type was None
    """
    # Generic loop for handling a source of key/value tuples (e.g., enumerate() or <dict>.items())
    def process_items_helper(key_value_iterator, data_map):
        made_change = False
        for key, value in key_value_iterator:
            if isinstance(value, str):
                altered_value = value
                for match in var_pattern.finditer(value):
                    var_type, var_loc, fallback_type, fallback_loc = _parse_var(match[1])
                    source_value = param_lookup_function(var_type, var_loc, fallback_type, fallback_loc)

                    # If we can't find a value, skip it
                    if source_value is None:
                        continue
                    try:
                        if altered_value == match[0]:           # complete replacement
                            altered_value = source_value
                        else:                                   # partial replacement

                            altered_value = altered_value.replace(match[0], source_value)
                    except TypeError:
                        pass
                if value != altered_value:
                    data_map[key] = altered_value
                    made_change = True
            elif isinstance(value, dict):
                if process_variables_loop(value, var_pattern, param_lookup_function):
                    made_change = True
            elif isinstance(value, list):
                if process_items_helper(enumerate(value), value):
                    made_change = True
        return made_change

    return process_items_helper(template.items(), template)


def _load_var_data(
    var_type,
    var_location,
    fallback_var_type,
    fallback_var_location,
    file_cache,
    self_data_source,
    service_file_directory
):
    """
Load data based on the type/path (see param_lookup_function parameter of process_variables for more info).

    :param var_type:        Either the type of the variable (see process_variables function) or None to
                            indicate that var_location is a raw value.
    :param var_location:    Either the location of the variable (see process_variables function) or a
                            raw value if var_type is None

    :return     None if the variable could not be resolved
    """
    value = None
    if var_type is None:
        value = var_location
    elif var_type == "self":
        value = _determine_variable_value_from_dict(self_data_source, var_location, None)
    elif var_type == "env":
        value = _determine_variable_value_from_dict(os.environ, var_location, None)
    elif var_type.startswith("file("):
        match = FILE_LOCATION_PATTERN.match(var_type)
        if match is None:
            return None
        data_source = _load_file_data(match[1], file_cache, service_file_directory)
        value = _determine_variable_value_from_dict(data_source, var_location, None)

    if value is None and fallback_var_location is not None:
        return _load_var_data(fallback_var_type, fallback_var_location, None, None,
                              file_cache, self_data_source, service_file_directory)
    return value


def _determine_variable_value_from_dict(source_dict, location_str, default):
    if location_str is None:
        return source_dict

    if default is not None:
        default = default.strip()

    # NOTE: String must be quoted to avoid issues with dashes and other reserved
    #       characters. If we just wrap the whole thing, dot separators won't work so:
    #       split and join with individually wrapped tokens.
    #         Original:  foo.bar-baz
    #         Wrapped:   "foo"."bar-baz"
    location = ".".join([f'"{token}"' for token in location_str.split(".")])
    source_value = jmespath.search(location, source_dict)
    if source_value is None:
        return default
    return source_value


def _self_var_data_lookup(group_dict, template):
    location = group_dict["loc"]
    default = group_dict.get("default")
    return _determine_variable_value_from_dict(template, location, default)


def _load_file_data(file_location, file_data_cache, service_file_directory):
    file_location = file_location.replace("~", str(Path.home()))
    file_location = file_location if os.path.isabs(file_location) else \
        os.path.join(service_file_directory, file_location)

    data = file_data_cache.get(file_location)
    if data is None:
        try:
            with open(file_location, "r") as f:
                if file_location.endswith(".json"):
                    data = json.load(f)
                elif file_location.endswith(".yml") or file_location.endswith(".yaml"):
                    data = yaml.safe_load(f)
        except Exception:
            data = {}
        file_data_cache[file_location] = data
    return data


def _token_to_type_and_loc(token: str) -> Tuple[Optional[str], Optional[str]]:
    file_match = FILE_LOCATION_PATTERN.match(token)
    if file_match is not None:
        if ":" not in token:
            return file_match[0], None

        return file_match[0].strip(), token[len(file_match[0]) + 1 :].strip()  # +1 for colon

    if ":" not in token:
        return None, token

    index = token.index(":")
    return token[:index].strip(), token[index + 1 :].strip()


def _parse_var(var_str):
    """
Returns a tuple of the var type, var loc, fallback type and fallback loc. See docs for the
param_lookup_function parameter of process_variables_loop for more info.
    """
    tokens = _tokenize_by_commas(var_str.strip())
    if not tokens:
        return None

    var_type, var_loc = _token_to_type_and_loc(tokens[0])

    if len(tokens) > 1:
        fallback_type, fallback_loc = _token_to_type_and_loc(tokens[1])
    else:
        fallback_type = None
        fallback_loc = None

    return var_type, var_loc, fallback_type, fallback_loc


def _tokenize_by_commas(string: str) -> Optional[List[str]]:
    """
    Tokenize the given value by commas, respecting quoted blocks.
    """
    if not string:
        return None

    quoted_comma_ranges = [range(m.start(0), m.end(0)) for m in QUOTED_WORD_SYNTAX.finditer(string)]

    def clean(s: str) -> str:
        s = s.strip()  # whitespace
        if len(s) > 0 and s[0] == '"' and s[len(s) - 1] == '"':  # surrounding quotes
            s = s[1:-1]
        if len(s) > 0 and s[0] == "'" and s[len(s) - 1] == "'":
            s = s[1:-1]
        return s

    block_start_index = 0
    search_start_index = block_start_index
    tokens = []
    index = string.find(",", search_start_index)
    while index > 0:
        is_quoted = False
        for quoted_comma_range in quoted_comma_ranges:
            if index in quoted_comma_range:
                is_quoted = True
                break
        if is_quoted:
            search_start_index = index + 1
        else:
            tokens.append(clean(string[block_start_index:index]))
            block_start_index = index + 1
            search_start_index = block_start_index
        index = string.find(",", search_start_index)

    if block_start_index < len(string):
        tokens.append(clean(string[block_start_index:]))
    return tokens



from __future__ import annotations
from collections import defaultdict

import json
import logging
import os
from typing import Dict, List, Tuple, Any
from charset_normalizer import from_fp

from checkov.terraform.context_parsers.registry import parser_registry
from checkov.terraform.plan_parser import parse_tf_plan, TF_PLAN_RESOURCE_ADDRESS
from checkov.common.runners.base_runner import filter_ignored_paths
from checkov.runner_filter import RunnerFilter


def create_definitions(
    root_folder: str,
    files: list[str] | None = None,
    runner_filter: RunnerFilter | None = None,
    out_parsing_errors: dict[str, str] | None = None,
) -> tuple[dict[str, dict[str, Any]], dict[str, list[tuple[int, str]]]]:
    runner_filter = runner_filter or RunnerFilter()
    out_parsing_errors = {} if out_parsing_errors is None else out_parsing_errors

    if root_folder:
        files = [] if not files else files
        for root, d_names, f_names in os.walk(root_folder):
            filter_ignored_paths(root, d_names, runner_filter.excluded_paths)
            filter_ignored_paths(root, f_names, runner_filter.excluded_paths)
            for file in f_names:
                file_ending = os.path.splitext(file)[1]
                if file_ending == '.json':
                    file_path = os.path.join(root, file)
                    try:
                        with open(file_path, "rb") as f:
                            try:
                                content = json.load(f)
                            except UnicodeDecodeError:
                                logging.debug(f"Encoding for file {file_path} is not UTF-8, trying to detect it")
                                content = str(from_fp(f).best())

                        if isinstance(content, dict) and content.get('terraform_version'):
                            files.append(file_path)
                    except Exception as e:
                        logging.debug(f'Failed to load json file {file_path}, skipping', stack_info=True)
                        out_parsing_errors[file_path] = str(e)

    tf_definitions = {}
    definitions_raw = {}
    if files:
        files = [os.path.realpath(file) for file in files]
        for file in files:
            if file.endswith(".json"):
                current_tf_definitions, current_definitions_raw = parse_tf_plan(file, out_parsing_errors)
                if current_tf_definitions and current_definitions_raw:
                    tf_definitions[file] = current_tf_definitions
                    definitions_raw[file] = current_definitions_raw
            else:
                logging.debug(f'Failed to load {file} as is not a .json file, skipping')
    return tf_definitions, definitions_raw


def build_definitions_context(
    definitions: dict[str, dict[str, list[dict[str, Any]]]],
    definitions_raw: Dict[str, List[Tuple[int, str]]]
) -> Dict[str, Dict[str, Any]]:
    definitions_context: dict[str, dict[str, Any]] = defaultdict(dict)
    supported_block_types = ("data", "resource")
    for full_file_path, definition in definitions.items():
        for block_type in supported_block_types:
            entities = definition.get(block_type, [])
            for entity in entities:
                context_parser = parser_registry.context_parsers[block_type]
                definition_path = context_parser.get_entity_context_path(entity)

                if len(definition_path) > 1:
                    resource_type = definition_path[0]
                    resource_name = definition_path[1]
                    entity_id = entity.get(resource_type, {}).get(resource_name, {}).get(TF_PLAN_RESOURCE_ADDRESS)
                else:
                    entity_id = definition_path[0]

                # Entity can exist only once per dir, for file as well
                entity_context = get_entity_context(
                    definitions=definitions,
                    definitions_raw=definitions_raw,
                    definition_path=definition_path,
                    full_file_path=full_file_path,
                    entity_id=entity_id,
                    block_type=block_type,
                )
                definitions_context[full_file_path][entity_id] = entity_context
    return definitions_context


def get_entity_context(
    definitions: dict[str, dict[str, list[dict[str, Any]]]],
    definitions_raw: dict[str, list[tuple[int, str]]],
    definition_path: list[str],
    full_file_path: str,
    entity_id: str,
    block_type: str = "resource",
) -> dict[str, Any]:
    entity_context: dict[str, Any] = {}

    if full_file_path not in definitions:
        logging.debug(
            f'Tried to look up file {full_file_path} in TF plan entity definitions, but it does not exist')
        return entity_context

    for resource in definitions.get(full_file_path, {}).get(block_type, []):
        resource_type = definition_path[0]
        resource_type_dict = resource.get(resource_type)
        if not resource_type_dict:
            continue
        resource_name = definition_path[1]
        resource_defintion = resource_type_dict.get(resource_name, {})
        if resource_defintion and resource_defintion.get(TF_PLAN_RESOURCE_ADDRESS) == entity_id:
            entity_context['start_line'] = resource_defintion['start_line'][0]
            entity_context['end_line'] = resource_defintion['end_line'][0]
            entity_context["code_lines"] = definitions_raw[full_file_path][
                entity_context["start_line"] : entity_context["end_line"]
            ]
            entity_context['address'] = resource_defintion[TF_PLAN_RESOURCE_ADDRESS]
            return entity_context
    return entity_context


def get_resource_id_without_nested_modules(address: str) -> str:
    """
    return resource id with the last module in the address
    example: from address='module.name1.module.name2.type.name' return 'module: module.name2.type.name'
    """
    return ".".join(address.split(".")[-4:])



from __future__ import annotations

from typing import Any

from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.terraform.checks.resource.base_resource_check import BaseResourceCheck

SUPPORTED_ENGINES = {
    "aurora",
    "aurora-mysql",
    "mysql"
}


class RDSClusterAuditLogging(BaseResourceCheck):
    def __init__(self) -> None:
        """
        NIST.800-53.r5 AC-2(4), NIST.800-53.r5 AC-4(26), NIST.800-53.r5 AC-6(9), NIST.800-53.r5 AU-10,
        NIST.800-53.r5 AU-12, NIST.800-53.r5 AU-2, NIST.800-53.r5 AU-3, NIST.800-53.r5 AU-6(3), NIST.800-53.r5 AU-6(4),
        NIST.800-53.r5 CA-7, NIST.800-53.r5 SC-7(10), NIST.800-53.r5 SC-7(9), NIST.800-53.r5 SI-3(8),
        NIST.800-53.r5 SI-4(20), NIST.800-53.r5 SI-7(8)
        Database logging should be enabled
        """
        name = "Ensure that RDS Cluster audit logging is enabled for MySQL engine"
        id = "CKV_AWS_325"
        supported_resources = ("aws_rds_cluster",)
        categories = (CheckCategories.LOGGING,)
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf: dict[str, list[Any]]) -> CheckResult:
        engine = conf.get("engine")
        if engine and isinstance(engine, list) and engine[0] not in SUPPORTED_ENGINES:
            # only MySQL cluster support easy audit logging export
            return CheckResult.UNKNOWN

        logs_exports = conf.get("enabled_cloudwatch_logs_exports")
        if (
            logs_exports
            and isinstance(logs_exports, list)
            and isinstance(logs_exports[0], list)
            and "audit" in logs_exports[0]
        ):
            return CheckResult.PASSED

        return CheckResult.FAILED


check = RDSClusterAuditLogging()



from checkov.common.models.enums import CheckResult, CheckCategories
from checkov.terraform.checks.resource.base_resource_check import BaseResourceCheck
from typing import List


class EKSControlPlaneLogging(BaseResourceCheck):
    def __init__(self):
        name = "Ensure Amazon EKS control plane logging enabled for all log types"
        id = "CKV_AWS_37"
        supported_resources = ['aws_eks_cluster']
        categories = [CheckCategories.KUBERNETES]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        """
            Looks for enabled_cluster_log_types at aws_eks_cluster:
            https://www.terraform.io/docs/providers/aws/r/eks_cluster.html
        :param conf: aws_eks_cluster configuration
        :return: <CheckResult>
        """
        log_types = ["api", "audit", "authenticator", "controllerManager", "scheduler"]
        if "enabled_cluster_log_types" in conf.keys() and conf["enabled_cluster_log_types"] and \
                conf["enabled_cluster_log_types"][0] is not None \
                and all(elem in conf["enabled_cluster_log_types"][0] for elem in log_types):
            return CheckResult.PASSED
        return CheckResult.FAILED

    def get_evaluated_keys(self) -> List[str]:
        return ['enabled_cluster_log_types']


check = EKSControlPlaneLogging()



from typing import List, Any
from checkov.common.models.enums import CheckCategories
from checkov.terraform.checks.resource.base_resource_negative_value_check import BaseResourceNegativeValueCheck


class SpringCloudAPIPortalPublicAccessIsDisabled(BaseResourceNegativeValueCheck):
    def __init__(self):
        name = "Ensures Spring Cloud API Portal Public Access Is Disabled"
        id = "CKV_AZURE_162"
        supported_resources = ['azurerm_spring_cloud_api_portal']
        categories = [CheckCategories.NETWORKING]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def get_inspected_key(self):
        return "public_network_access_enabled"

    def get_forbidden_values(self) -> List[Any]:
        return [True]


check = SpringCloudAPIPortalPublicAccessIsDisabled()



from __future__ import annotations

from typing import Any

from checkov.terraform.checks.resource.base_resource_check import BaseResourceCheck
from checkov.common.models.enums import CheckResult, CheckCategories

ALLOWED_ACCESS_TYPES = (
    "user_by_email",
    "group_by_email",
    "domain",
    "view",
    "routine",
    "dataset",
)


class GoogleBigQueryDatasetPublicACL(BaseResourceCheck):
    def __init__(self) -> None:
        name = "Ensure that BigQuery datasets are not anonymously or publicly accessible"
        id = "CKV_GCP_15"
        supported_resources = ("google_bigquery_dataset",)
        categories = (CheckCategories.GENERAL_SECURITY,)
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf: dict[str, list[Any]]) -> CheckResult:
        """
            Looks for ACL configuration at bigquery_dataset:
            https://www.terraform.io/docs/providers/google/r/bigquery_dataset.html#access
        :param conf: bigquery_dataset configuration
        :return: <CheckResult>
        """
        if "access" in conf.keys():
            for access in conf["access"]:
                if "special_group" in access:
                    if access["special_group"] in [["allAuthenticatedUsers"], ["allUsers"]]:
                        self.evaluated_keys = [f'access/[{conf["access"].index(access)}]/special_group']
                        return CheckResult.FAILED
                # access block with only the role key found in the statefile
                # when manually adding "allUsers" to the dataset
                elif not any(key in access for key in ALLOWED_ACCESS_TYPES):
                    self.evaluated_keys = [f'access/[{conf["access"].index(access)}]']
                    return CheckResult.FAILED
            self.evaluated_keys = ['access']
        return CheckResult.PASSED


check = GoogleBigQueryDatasetPublicACL()



from checkov.common.models.enums import CheckCategories
from checkov.terraform.checks.resource.gcp.AbsGooglePostgresqlDatabaseFlags import AbsGooglePostgresqlDatabaseFlags

FLAG_NAME = 'log_statement'
FLAG_VALUES = [
    'ddl',
    'mod',
    'all'
]


class GoogleCloudPostgreSqlLogStatement(AbsGooglePostgresqlDatabaseFlags):
    def __init__(self):
        name = "Ensure GCP PostgreSQL logs SQL statements"
        check_id = "CKV_GCP_111"
        supported_resources = ['google_sql_database_instance']
        categories = [CheckCategories.LOGGING]
        super().__init__(
            name=name,
            id=check_id,
            categories=categories,
            supported_resources=supported_resources,
            flag_name=FLAG_NAME,
            flag_values=FLAG_VALUES
        )


check = GoogleCloudPostgreSqlLogStatement()



from __future__ import annotations

import logging
import os
from copy import deepcopy
from typing import Optional, Dict, Mapping, Set, Tuple, Callable, Any, List, cast, TYPE_CHECKING

import deep_merge

from checkov.common.runners.base_runner import filter_ignored_paths, IGNORE_HIDDEN_DIRECTORY_ENV
from checkov.common.util.consts import DEFAULT_EXTERNAL_MODULES_DIR, RESOLVED_MODULE_ENTRY_NAME
from checkov.common.util.type_forcers import force_list
from checkov.common.variables.context import EvaluationContext
from checkov.terraform.graph_builder.graph_components.block_types import BlockType
from checkov.terraform.graph_builder.graph_components.module import Module
from checkov.terraform.module_loading.content import ModuleContent
from checkov.terraform.module_loading.module_finder import load_tf_modules
from checkov.terraform.module_loading.registry import module_loader_registry as default_ml_registry, \
    ModuleLoaderRegistry
from checkov.common.util.parser_utils import is_acceptable_module_param
from checkov.terraform.modules.module_utils import load_or_die_quietly, safe_index, \
    remove_module_dependency_from_path, \
    clean_parser_types, serialize_definitions
from checkov.terraform.modules.module_objects import TFModule, TFDefinitionKey

if TYPE_CHECKING:
    from typing_extensions import TypeGuard


def _filter_ignored_paths(root: str, paths: list[str], excluded_paths: list[str] | None) -> None:
    filter_ignored_paths(root, paths, excluded_paths)
    for path in force_list(paths):
        if path == default_ml_registry.external_modules_folder_name:
            paths.remove(path)


class TFParser:
    def __init__(self, module_class: type[Module] = Module) -> None:
        self.module_class = module_class
        self._parsed_directories: set[str] = set()
        self.external_modules_source_map: Dict[Tuple[str, str], str] = {}
        self.module_address_map: Dict[Tuple[str, str], str] = {}
        self.loaded_files_map: dict[str, dict[str, list[dict[str, Any]]] | None] = {}
        self.external_variables_data: list[tuple[str, Any, str]] = []

    def _init(self, directory: str,
              out_evaluations_context: Dict[str, Dict[str, EvaluationContext]] | None,
              out_parsing_errors: Dict[str, Exception] | None,
              env_vars: Mapping[str, str] | None,
              download_external_modules: bool,
              external_modules_download_path: str,
              excluded_paths: Optional[List[str]] = None,
              tf_var_files: Optional[List[str]] = None) -> None:
        self.directory = directory
        self.out_definitions: dict[TFDefinitionKey, dict[str, list[dict[str, Any]]]] = {}
        self.out_evaluations_context = {} if out_evaluations_context is None else out_evaluations_context
        self.out_parsing_errors = {} if out_parsing_errors is None else out_parsing_errors
        self.env_vars = dict(os.environ) if env_vars is None else env_vars
        self.download_external_modules = download_external_modules
        self.external_modules_download_path = external_modules_download_path
        self.external_modules_source_map = {}
        self.module_address_map = {}
        self.tf_var_files = tf_var_files
        self.dirname_cache: dict[str, str] = {}
        self.excluded_paths = excluded_paths
        self.visited_definition_keys: set[TFDefinitionKey] = set()
        self.module_to_resolved: dict[tuple[TFDefinitionKey | None, str], list[TFDefinitionKey]] = {}
        self.keys_to_remove: set[TFDefinitionKey] = set()

    def _check_process_dir(self, directory: str) -> bool:
        if directory not in self._parsed_directories:
            self._parsed_directories.add(directory)
            return True
        else:
            return False

    def parse_directory(
        self,
        directory: str,
        out_evaluations_context: Dict[str, Dict[str, EvaluationContext]] | None = None,
        out_parsing_errors: Dict[str, Exception] | None = None,
        env_vars: Mapping[str, str] | None = None,
        download_external_modules: bool = False,
        external_modules_download_path: str = DEFAULT_EXTERNAL_MODULES_DIR,
        excluded_paths: Optional[List[str]] = None,
        vars_files: Optional[List[str]] = None,
        external_modules_content_cache: Optional[Dict[str, ModuleContent | None]] = None,
    ) -> dict[TFDefinitionKey, dict[str, list[dict[str, Any]]]]:
        self._init(directory, out_evaluations_context, out_parsing_errors, env_vars,
                   download_external_modules, external_modules_download_path, excluded_paths)
        self._parsed_directories.clear()
        default_ml_registry.root_dir = directory
        default_ml_registry.download_external_modules = download_external_modules
        default_ml_registry.external_modules_folder_name = external_modules_download_path
        default_ml_registry.module_content_cache = external_modules_content_cache if external_modules_content_cache else {}
        load_tf_modules(directory)
        self._parse_directory(dir_filter=lambda d: self._check_process_dir(d), vars_files=vars_files)
        self._update_resolved_modules()
        return self.out_definitions

    def parse_file(self, file: str, parsing_errors: dict[str, Exception]) -> Optional[Dict[str, Any]]:
        if file.endswith(".tf") or file.endswith(".tf.json") or file.endswith(".hcl"):
            parse_result = load_or_die_quietly(file, parsing_errors)
            if parse_result:
                parse_result = serialize_definitions(parse_result)
                parse_result = clean_parser_types(parse_result)
                return parse_result

        return None

    def _parse_directory(self, include_sub_dirs: bool = True,
                         module_loader_registry: ModuleLoaderRegistry = default_ml_registry,
                         dir_filter: Callable[[str], bool] = lambda _: True,
                         vars_files: Optional[List[str]] = None) -> None:

        keys_referenced_as_modules: set[TFDefinitionKey] = set()

        if include_sub_dirs:
            for sub_dir, d_names, _ in os.walk(self.directory):
                _filter_ignored_paths(sub_dir, d_names, self.excluded_paths)
                if dir_filter(os.path.abspath(sub_dir)):
                    self._internal_dir_load(sub_dir, module_loader_registry, dir_filter,
                                            keys_referenced_as_modules, vars_files=vars_files,
                                            excluded_paths=self.excluded_paths)
        else:
            self._internal_dir_load(self.directory, module_loader_registry, dir_filter,
                                    keys_referenced_as_modules, vars_files=vars_files)

        for key in keys_referenced_as_modules:
            if key in self.out_definitions:
                del self.out_definitions[key]

    def _internal_dir_load(
        self,
        directory: str,
        module_loader_registry: ModuleLoaderRegistry,
        dir_filter: Callable[[str], bool],
        keys_referenced_as_modules: set[TFDefinitionKey],
        specified_vars: Optional[Mapping[str, str]] = None,
        vars_files: Optional[List[str]] = None,
        excluded_paths: Optional[List[str]] = None,
        nested_modules_data: dict[str, Any] | None = None,
    ) -> None:

        dir_contents = list(os.scandir(directory))
        if excluded_paths or IGNORE_HIDDEN_DIRECTORY_ENV:
            filter_ignored_paths(directory, dir_contents, excluded_paths)

        tf_files_to_load = self.handle_variables(dir_contents, vars_files, specified_vars)
        files_to_data = self._load_files(tf_files_to_load)
        for file, data in sorted(files_to_data, key=lambda x: x[0]):
            if not data:
                continue
            self.out_definitions[TFDefinitionKey(file)] = data
            self.add_external_vars_from_data(data, file)

        force_final_module_load = False
        for i in range(0, 10):
            logging.debug(f"Module load loop {i}")
            dir_filter(directory)
            has_more_modules = self._load_modules(
                directory, module_loader_registry, dir_filter,
                keys_referenced_as_modules, force_final_module_load,
                nested_modules_data=nested_modules_data
            )
            made_var_changes = False
            if not has_more_modules:
                break
            elif not made_var_changes:
                force_final_module_load = True

    def _load_files(
        self,
        files: list[os.DirEntry[str]],
    ) -> list[tuple[str, dict[str, list[dict[str, Any]]] | None]]:
        def _load_file(
            file: os.DirEntry[str]
        ) -> tuple[tuple[str, dict[str, list[dict[str, Any]]] | None], dict[str, Exception]]:
            parsing_errors: dict[str, Exception] = {}
            result = load_or_die_quietly(file, parsing_errors)
            for path, e in parsing_errors.items():
                parsing_errors[path] = e

            return (file.path, result), parsing_errors

        files_to_data: list[tuple[str, dict[str, list[dict[str, Any]]] | None]] = []
        files_to_parse = []
        for file in files:
            data = self.loaded_files_map.get(file.path)
            if data:
                files_to_data.append((file.path, data))
            else:
                files_to_parse.append(file)

        results = [_load_file(f) for f in files_to_parse]
        for result, parsing_errors in results:
            self.out_parsing_errors.update(parsing_errors)
            files_to_data.append(result)
            if result[0] not in self.loaded_files_map:
                self.loaded_files_map[result[0]] = result[1]
        return files_to_data

    def _load_modules(self, root_dir: str, module_loader_registry: ModuleLoaderRegistry,
                      dir_filter: Callable[[str], bool],
                      keys_referenced_as_modules: Set[TFDefinitionKey], ignore_unresolved_params: bool = False,
                      nested_modules_data: dict[str, Any] | None = None) -> bool:
        all_module_definitions: dict[TFDefinitionKey, dict[str, list[dict[str, Any]]]] = {}
        skipped_a_module = False
        for file in list(self.out_definitions.keys()):
            if not self.should_loaded_file(file, root_dir):
                continue

            #  Dont run over the nested because we already run on them - dont remove.
            if file.tf_source_modules:
                continue

            file_data = self.out_definitions.get(file)
            if file_data is None:
                continue
            module_calls = file_data.get("module")
            if not module_calls or not isinstance(module_calls, list):
                continue

            for module_call in module_calls:
                if not isinstance(module_call, dict):
                    continue

                for module_call_name, module_call_data in module_call.items():
                    if not isinstance(module_call_data, dict):
                        continue

                    file_key = self.get_file_key_with_nested_data(file, nested_modules_data)
                    current_nested_data = (file_key, module_call_name)
                    resolved_loc_list = []
                    if current_nested_data in self.module_to_resolved:
                        resolved_loc_list = self.module_to_resolved[current_nested_data]
                    self.module_to_resolved[current_nested_data] = resolved_loc_list

                    specified_vars = {k: v[0] if isinstance(v, list) else v for k, v in module_call_data.items()
                                      if k != "source" and k != "version"}
                    skipped_a_module = self.should_skip_a_module(specified_vars, ignore_unresolved_params)
                    if skipped_a_module:
                        continue

                    version = self.get_module_version(module_call_data)
                    source = self.get_module_source(module_call_data, module_call_name, file)
                    if not source:
                        continue

                    try:
                        content_path = self.get_content_path(module_loader_registry, root_dir, source, version)
                        if not content_path:
                            continue
                        new_nested_modules_data = {'module_name': module_call_name, 'file': file, 'nested_modules_data': nested_modules_data}
                        self._internal_dir_load(
                            directory=content_path,
                            module_loader_registry=module_loader_registry,
                            dir_filter=dir_filter, specified_vars=specified_vars,
                            keys_referenced_as_modules=keys_referenced_as_modules,
                            nested_modules_data=new_nested_modules_data
                        )

                        module_definitions = {
                            path: definition
                            for path, definition in self.out_definitions.items()
                            if self.get_dirname(path) == content_path and not path.tf_source_modules
                        }
                        if not module_definitions:
                            continue

                        keys = list(module_definitions.keys())
                        for key in keys:
                            if not self.should_process_key(key, file):
                                continue
                            keys_referenced_as_modules.add(key)
                            new_key = self.get_new_nested_module_key(key, file, module_call_name, nested_modules_data)
                            if new_key in self.visited_definition_keys:
                                del module_definitions[key]
                                del self.out_definitions[key]
                                continue

                            module_definitions[new_key] = module_definitions[key]
                            del module_definitions[key]
                            del self.out_definitions[key]
                            self.keys_to_remove.add(key)

                            self.visited_definition_keys.add(new_key)
                            if new_key not in resolved_loc_list:
                                resolved_loc_list.append(new_key)

                        if all_module_definitions:
                            deep_merge.merge(all_module_definitions, module_definitions)
                        else:
                            all_module_definitions = module_definitions

                        self.external_modules_source_map[(source, version)] = content_path
                    except Exception as e:
                        logging.warning(f"Unable to load module - source: {source}, version: {version}, error: {str(e)}")

        if all_module_definitions:
            deep_merge.merge(self.out_definitions, all_module_definitions)
        return skipped_a_module

    def parse_hcl_module(
        self,
        source_dir: str,
        source: str,
        download_external_modules: bool = False,
        external_modules_download_path: str = DEFAULT_EXTERNAL_MODULES_DIR,
        parsing_errors: dict[str, Exception] | None = None,
        excluded_paths: list[str] | None = None,
        vars_files: list[str] | None = None,
        external_modules_content_cache: dict[str, ModuleContent | None] | None = None,
        create_graph: bool = True,
    ) -> tuple[Module | None, dict[TFDefinitionKey, dict[str, Any]]]:
        tf_definitions = self.parse_directory(
            directory=source_dir, out_evaluations_context={},
            out_parsing_errors=parsing_errors if parsing_errors is not None else {},
            download_external_modules=download_external_modules,
            external_modules_download_path=external_modules_download_path, excluded_paths=excluded_paths,
            vars_files=vars_files, external_modules_content_cache=external_modules_content_cache
        )
        tf_definitions = clean_parser_types(tf_definitions)
        tf_definitions = serialize_definitions(tf_definitions)

        module = None
        if create_graph:
            module, tf_definitions = self.parse_hcl_module_from_tf_definitions(tf_definitions, source_dir, source)

        return module, tf_definitions

    def _remove_unused_path_recursive(self, path: TFDefinitionKey) -> None:
        self.out_definitions.pop(path, None)
        for key in list(self.module_to_resolved.keys()):
            file_key = None
            if isinstance(key[0], TFDefinitionKey):
                file_key = key[0]
            elif key[0] is not None:
                file_key, module_index, module_name = key
            if path == file_key:
                for resolved_path in self.module_to_resolved[key]:
                    self._remove_unused_path_recursive(resolved_path)
                self.module_to_resolved.pop(key, None)

    def _update_resolved_modules(self) -> None:
        for key in list(self.module_to_resolved.keys()):
            file_key, module_name = key
            if file_key in self.keys_to_remove:
                for path in self.module_to_resolved[key]:
                    self._remove_unused_path_recursive(path)
                self.module_to_resolved.pop(key, None)

        for key, resolved_list in self.module_to_resolved.items():
            file_key, module_name = key
            if file_key not in self.out_definitions:
                continue

            idx = self.get_idx_by_module_name(self.out_definitions[file_key]['module'], module_name)
            if idx is None:
                continue

            self.out_definitions[file_key]['module'][idx][module_name][RESOLVED_MODULE_ENTRY_NAME] = resolved_list

    @staticmethod
    def get_idx_by_module_name(module_data_list: list[dict[str, Any]], module_name: str) -> int | None:
        for idx, module_data in enumerate(module_data_list):
            if module_name in module_data:
                return idx

        return None

    def parse_hcl_module_from_tf_definitions(
        self,
        tf_definitions: Dict[TFDefinitionKey, Dict[str, Any]],
        source_dir: str,
        source: str,
    ) -> Tuple[Module, Dict[TFDefinitionKey, Dict[str, Any]]]:
        module = self.get_new_module(
            source_dir=source_dir,
            module_address_map=self.module_address_map,
            external_modules_source_map=self.external_modules_source_map,
        )
        self.add_tfvars(module, source)
        copy_of_tf_definitions = deepcopy(tf_definitions)
        for file_path, blocks in copy_of_tf_definitions.items():
            for block_type in blocks:
                try:
                    module.add_blocks(block_type, blocks[block_type], file_path, source)
                except Exception as e:
                    logging.warning(f'Failed to add block {blocks[block_type]}. Error:')
                    logging.warning(e, exc_info=False)
        return module, tf_definitions

    def get_file_key_with_nested_data(
        self, file: TFDefinitionKey | None, nested_data: dict[str, Any] | None
    ) -> TFDefinitionKey | None:
        if not nested_data or file is None:
            return file
        nested_str = self.get_file_key_with_nested_data(nested_data.get("file"), nested_data.get('nested_modules_data'))
        nested_module_name = nested_data.get('module_name')
        return get_tf_definition_object_from_module_dependency(file, nested_str, nested_module_name)

    def get_new_nested_module_key(
        self, key: TFDefinitionKey, file: TFDefinitionKey | None, module_name: str | None, nested_data: Optional[dict[str, Any]]
    ) -> TFDefinitionKey:
        if not nested_data or not file:
            return get_tf_definition_object_from_module_dependency(key, file, module_name)
        visited_key_to_add = get_tf_definition_object_from_module_dependency(key, file, module_name)
        self.visited_definition_keys.add(visited_key_to_add)
        nested_key = self.get_new_nested_module_key(file, nested_data.get('file'),
                                                    nested_data.get('module_name'),
                                                    nested_data.get('nested_modules_data'))
        return get_tf_definition_object_from_module_dependency(key, nested_key, module_name)

    def add_tfvars(self, module: Module, source: str) -> None:
        if not self.external_variables_data:
            return
        for (var_name, default, path) in self.external_variables_data:
            if ".tfvars" in path:
                block = [{var_name: {"default": default}}]
                module.add_blocks(BlockType.TF_VARIABLE, block, path, source)

    def get_dirname(self, path: str | TFDefinitionKey) -> str:
        if isinstance(path, TFDefinitionKey):
            path = path.file_path
        dirname_path = self.dirname_cache.get(path)
        if not dirname_path:
            dirname_path = os.path.dirname(path)
            self.dirname_cache[path] = dirname_path
        return dirname_path

    def should_loaded_file(self, file: TFDefinitionKey, root_dir: str) -> bool:
        return not self.get_dirname(file) != root_dir

    def get_module_source(
        self, module_call_data: dict[str, Any], module_call_name: str, file: str | TFDefinitionKey
    ) -> Optional[str]:
        source = module_call_data.get("source")
        if not source or not isinstance(source, list):
            return None
        source = source[0]
        if not self.is_valid_source(source, module_call_name):
            return None

        if source.startswith("./") or source.startswith("../"):
            file_to_load = file.file_path if isinstance(file, TFDefinitionKey) else file
            source = os.path.normpath(os.path.join(os.path.dirname(remove_module_dependency_from_path(file_to_load)), source))
        return source

    def add_external_vars_from_data(self, data: dict[str, Any], file: str) -> None:
        var_blocks = data.get("variable")
        if var_blocks and isinstance(var_blocks, list):
            for var_block in var_blocks:
                if not isinstance(var_block, dict):
                    continue
                for var_name, var_definition in var_block.items():
                    if not isinstance(var_definition, dict):
                        continue

                    default_value = var_definition.get("default")
                    if default_value is not None and isinstance(default_value, list):
                        self.external_variables_data.append((var_name, default_value[0], file))

    def handle_variables(
        self,
        dir_contents: list[os.DirEntry[str]],
        vars_files: None | list[str],
        specified_vars: Mapping[str, str] | None,
    ) -> list[os.DirEntry[str]]:
        tf_files_to_load = []
        hcl_tfvars: Optional[os.DirEntry[str]] = None
        json_tfvars: Optional[os.DirEntry[str]] = None
        auto_vars_files: List[os.DirEntry[str]] = []
        explicit_var_files: List[os.DirEntry[str]] = []
        for file in dir_contents:
            try:
                if not file.is_file():
                    continue
            except OSError:
                continue

            if file.name == "terraform.tfvars.json":
                json_tfvars = file
            elif file.name == "terraform.tfvars":
                hcl_tfvars = file
            elif file.name.endswith(".auto.tfvars.json") or file.name.endswith(".auto.tfvars"):
                auto_vars_files.append(file)
            elif vars_files and file.path in vars_files:
                explicit_var_files.append(file)
            elif file.name.endswith(".tf") or file.name.endswith('.hcl'):  # TODO: add support for .tf.json
                tf_files_to_load.append(file)

        for key, value in self.env_vars.items():
            if not key.startswith("TF_VAR_"):
                continue
            self.external_variables_data.append((key[7:], value, f"env:{key}"))
        if hcl_tfvars:  # terraform.tfvars
            data = load_or_die_quietly(hcl_tfvars, self.out_parsing_errors, clean_definitions=False)
            if data:
                self.external_variables_data.extend([(k, safe_index(v, 0), hcl_tfvars.path) for k, v in data.items()])
        if json_tfvars:  # terraform.tfvars.json
            data = load_or_die_quietly(json_tfvars, self.out_parsing_errors)
            if data:
                self.external_variables_data.extend([(k, v, json_tfvars.path) for k, v in data.items()])

        auto_var_files_to_data = self._load_files(auto_vars_files)
        for var_file, data in sorted(auto_var_files_to_data, key=lambda x: x[0]):
            if data:
                self.external_variables_data.extend([(k, v, var_file) for k, v in data.items()])

        explicit_var_files_to_data = self._load_files(explicit_var_files)
        # it's possible that os.scandir returned the var files in a different order than they were specified
        if vars_files:
            for var_file, data in sorted(explicit_var_files_to_data, key=lambda x: vars_files.index(x[0])):  # type:ignore[union-attr]  # false-positive
                if data:
                    self.external_variables_data.extend([(k, v, var_file) for k, v in data.items()])

        if specified_vars:  # specified
            self.external_variables_data.extend([(k, v, "manual specification") for k, v in specified_vars.items()])

        return tf_files_to_load

    @staticmethod
    def get_module_version(module_call_data: dict[str, Any]) -> str:
        version = module_call_data.get("version", "latest")
        if version and isinstance(version, list):
            version = version[0]
        return cast(str, version)

    @staticmethod
    def should_process_key(key: TFDefinitionKey, file: TFDefinitionKey) -> bool:
        return bool(not key.tf_source_modules or file.tf_source_modules)

    @staticmethod
    def is_valid_source(source: Any, module_call_name: str) -> TypeGuard[str]:
        if not isinstance(source, str):
            logging.debug(f"Skipping loading of {module_call_name} as source is not a string, it is: {source}")
            return False
        elif source in ['./', '.']:
            logging.debug(f"Skipping loading of {module_call_name} as source is the current dir")
            return False
        return True

    @staticmethod
    def should_skip_a_module(specified_vars: dict[str, Any], ignore_unresolved_params: bool) -> bool:
        if not ignore_unresolved_params:
            has_unresolved_params = False
            for k, v in specified_vars.items():
                if not is_acceptable_module_param(v) or not is_acceptable_module_param(k):
                    has_unresolved_params = True
                    break
            if has_unresolved_params:
                return True
        return False

    @staticmethod
    def get_content_path(module_loader_registry: ModuleLoaderRegistry, root_dir: str, source: str, version: str) -> Optional[str]:
        content = module_loader_registry.load(root_dir, source, version)
        if not content or not content.loaded():
            logging.info(f'Got no content for {source}:{version}')
            return None
        return content.path()

    @staticmethod
    def get_new_module(
            source_dir: str,
            module_address_map: dict[tuple[str, str], str],
            external_modules_source_map: dict[tuple[str, str], str],
    ) -> Module:
        return Module(
            source_dir=source_dir,
            module_address_map=module_address_map,
            external_modules_source_map=external_modules_source_map,
        )


def is_nested_object(full_path: TFDefinitionKey) -> bool:
    return True if full_path.tf_source_modules else False


def get_tf_definition_object_from_module_dependency(
    path: TFDefinitionKey, module_dependency: TFDefinitionKey | None, module_dependency_name: str | None
) -> TFDefinitionKey:
    if not module_dependency:
        return path
    if not is_nested_object(module_dependency):
        return TFDefinitionKey(path.file_path, TFModule(path=module_dependency.file_path, name=module_dependency_name))
    return TFDefinitionKey(path.file_path, TFModule(path=module_dependency.file_path, name=module_dependency_name, nested_tf_module=module_dependency.tf_source_modules))



from checkov.terraform.checks.resource.base_resource_value_check import BaseResourceValueCheck
from checkov.common.models.enums import CheckCategories, CheckResult


class RedshiftClusterUseEnhancedVPCRouting(BaseResourceValueCheck):
    def __init__(self):
        """
        NIST.800-53.r5 AC-4, NIST.800-53.r5 AC-4(21), NIST.800-53.r5 SC-7,
        NIST.800-53.r5 SC-7(11), NIST.800-53.r5 SC-7(20),
        NIST.800-53.r5 SC-7(21), NIST.800-53.r5 SC-7(4), NIST.800-53.r5 SC-7(9)
        Redshift clusters should use enhanced VPC routing
        """
        name = "Ensure Redshift clusters use enhanced VPC routing"
        id = "CKV_AWS_321"
        supported_resources = ['aws_redshift_cluster']
        categories = [CheckCategories.GENERAL_SECURITY]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources,
                         missing_block_result=CheckResult.FAILED)

    def get_inspected_key(self):
        return "enhanced_vpc_routing"


check = RedshiftClusterUseEnhancedVPCRouting()



from __future__ import annotations

import logging
import copy
from collections import defaultdict
from typing import TYPE_CHECKING, Dict, List, Optional
from typing_extensions import TypedDict
from checkov.common.util.secrets import omit_secret_value_from_line
from checkov.secrets.git_types import EnrichedPotentialSecretMetadata, EnrichedPotentialSecret, Commit, ADDED, REMOVED, \
    GIT_HISTORY_OPTIONS, CommitDiff, GIT_HISTORY_NOT_BEEN_REMOVED

if TYPE_CHECKING:
    from detect_secrets.core.potential_secret import PotentialSecret

RENAME_STR = 'rename'
FILE_RESULTS_STR = 'file_results'
RAW_STORE_TYPES = {RENAME_STR, FILE_RESULTS_STR}


class RawStore(TypedDict):
    file_results: List[PotentialSecret]
    file_name: str
    commit: Commit
    type: str  # rename / file results
    rename_from: str
    rename_to: str


class GitHistorySecretStore:
    def __init__(self) -> None:
        self.secrets_by_file_value_type: Dict[str, List[EnrichedPotentialSecret]] = {}

    def set_secret_map(self, file_results: List[PotentialSecret], file_name: str, commit: Commit) -> None:
        # First find if secret was moved in the file
        equal_secret_in_commit: Dict[str, List[str]] = defaultdict(list)
        for secret in file_results:
            secret_key = get_secret_key(file_name, secret.secret_hash, secret.type)
            equal_secret_in_commit[secret_key].append(ADDED if secret.is_added else REMOVED)

        for secret in file_results:
            if not secret.filename or 'None' in secret.filename:
                secret.filename = file_name
            secret_key = get_secret_key(file_name, secret.secret_hash, secret.type)
            if all(value in equal_secret_in_commit[secret_key] for value in GIT_HISTORY_OPTIONS):
                continue
            if secret.is_added:
                self._add_new_secret(secret_key, secret, commit)
            if secret.is_removed:
                self._update_removed_secret(secret_key, secret, file_name, commit)

    def _add_new_secret(self, secret_key: str, secret: PotentialSecret, commit: Commit) -> None:
        if secret_key not in self.secrets_by_file_value_type:
            self.secrets_by_file_value_type[secret_key] = []
        else:
            all_removed = all(
                potential_secret.get('removed_commit_hash') for potential_secret in
                self.secrets_by_file_value_type[secret_key])
            # Update secret map with the new potential secret
            if all_removed:
                self.secrets_by_file_value_type[secret_key][0].update({'potential_secret': secret,
                                                                       'removed_commit_hash': '',
                                                                       'removed_date': ''})
                return
        code_line = search_for_code_line(commit.files[secret.filename], secret.secret_value, secret.is_added)
        enriched_potential_secret: EnrichedPotentialSecret = {
            'added_commit_hash': commit.metadata.commit_hash,
            'removed_commit_hash': '',
            'potential_secret': secret,
            'code_line': code_line,
            'added_by': commit.metadata.committer,
            'removed_date': '',
            'added_date': commit.metadata.committed_datetime
        }
        self.secrets_by_file_value_type[secret_key].append(enriched_potential_secret)

    def _update_removed_secret(self, secret_key: str, secret: PotentialSecret, file_name: str, commit: Commit) -> None:
        # Try to find the corresponding added secret in the git history secret map
        secrets_in_file = self.secrets_by_file_value_type.get(secret_key, None)
        if secrets_in_file:
            for secret_in_file in secrets_in_file:
                if secret_in_file['potential_secret'].is_added:
                    secret_in_file['removed_commit_hash'] = commit.metadata.commit_hash
                    secret_in_file['potential_secret'] = secret
                    secret_in_file['removed_date'] = commit.metadata.committed_datetime
                    break
        else:
            logging.warning(f"No added secret commit found for secret in file {file_name}.")

    def handle_renamed_file(self, rename_from: str, rename_to: str, commit: Commit) -> None:
        temp_secrets_by_file_value_type: Dict[str, List[EnrichedPotentialSecret]] = {}
        for secret_key in self.secrets_by_file_value_type.keys():
            if rename_from in secret_key:
                new_secret_key = secret_key.replace(rename_from, rename_to)
                temp_secrets_by_file_value_type[new_secret_key] = []
                secret_in_file = self.secrets_by_file_value_type[secret_key]
                for secret_data in secret_in_file:
                    # defines the secret in the old file as removed and add the secret to the new file
                    secret_data['removed_commit_hash'] = commit.metadata.commit_hash
                    secret_data['removed_date'] = commit.metadata.committed_datetime
                    new_secret = copy.deepcopy(secret_data['potential_secret'])
                    new_secret.filename = rename_to
                    code = secret_data.get('code_line')
                    enriched_potential_secret: EnrichedPotentialSecret = {
                        'added_commit_hash': commit.metadata.commit_hash,
                        'removed_commit_hash': '',
                        'potential_secret': new_secret,
                        'code_line': code,
                        'added_by': secret_data.get('added_by'),
                        'removed_date': '',
                        'added_date': secret_data.get('added_date')
                    }
                    temp_secrets_by_file_value_type[new_secret_key].append(enriched_potential_secret)
        self.secrets_by_file_value_type.update(temp_secrets_by_file_value_type)

    def get_added_and_removed_commit_hash(self, key: str, secret: PotentialSecret, root_folder: Optional[str]) -> EnrichedPotentialSecretMetadata:
        """
        now we have only the current commit_hash - in the added_commit_hash or in the removed_commit_hash.
        in the next step we will add the connection and the missing data
        The key is built like this:
        '{added_commit_hash}_{removed_commit_hash or the string GIT_HISTORY_NOT_BEEN_REMOVED
        if the secret not been removed}_{file_name}'
        returns (added, removed, code)
        """
        try:
            secret_key = get_secret_key(secret.filename, secret.secret_hash, secret.type)  # by value type
            enriched_secrets: List[EnrichedPotentialSecret] = self.secrets_by_file_value_type.get(secret_key, [])
            if not enriched_secrets and root_folder:
                # sometimes the secret key is from the project path instead of abs path
                filename = f'{root_folder}/{secret.filename}'
                secret_key = get_secret_key(filename, secret.secret_hash, secret.type)  # by value type
                enriched_secrets = self.secrets_by_file_value_type.get(secret_key, [])
                if not enriched_secrets:
                    logging.warning(f'Did not find added_commit_hash and removed_commit_hash for {secret_key}')
                    return {}
            chosen_secret = enriched_secrets[0]
            if len(enriched_secrets) > 1:
                res = key.split("_")
                added, removed = res[0], res[1]
                if removed == GIT_HISTORY_NOT_BEEN_REMOVED:
                    removed = ''
                for enriched_secret in enriched_secrets:
                    if added == enriched_secret.get('added_commit_hash') and \
                            removed == enriched_secret.get('removed_commit_hash'):
                        chosen_secret = enriched_secret
                        break

            return {
                'added_commit_hash': chosen_secret.get('added_commit_hash', ''),
                'removed_commit_hash': chosen_secret.get('removed_commit_hash', ''),
                'code_line': chosen_secret.get('code_line'),
                'added_by': chosen_secret.get('added_by'),
                'removed_date': chosen_secret.get('removed_date'),
                'added_date': chosen_secret.get('added_date')
            }
        except Exception as e:
            logging.warning(f"Failed set added_commit_hash and removed_commit_hash due to: {str(e)}")
            return {}


def search_for_code_line(commit_diff: CommitDiff, secret_value: Optional[str], is_added: Optional[bool]) -> str:
    if not commit_diff:
        logging.warning(f'missing file name for {commit_diff}, hence no available code line')
    if secret_value is None:
        return ''
    splitted = commit_diff.split('\n')
    start_char = '+' if is_added else '-'
    for line in splitted:
        if line.startswith(start_char) and secret_value in line:
            # remove +/- in the beginning & spaces and omit
            return omit_secret_value_from_line(secret_value, line[1:].strip()) or ''
    return ''  # not found


def get_secret_key(file_name: str, secret_hash: str, secret_type: str) -> str:
    """
    One way to create a secret key for the secret map
    """
    secret_key = f'{file_name}_{secret_hash}_{secret_type.replace(" ", "-")}'
    return secret_key



from typing import Dict, List, Any

from checkov.common.models.enums import CheckResult, CheckCategories
from checkov.terraform.checks.resource.base_resource_check import BaseResourceCheck


class ECSContainerReadOnlyRoot(BaseResourceCheck):
    def __init__(self) -> None:
        """
        NIST.800-53.r5 AC-2(1), NIST.800-53.r5 AC-3, NIST.800-53.r5 AC-3(15), NIST.800-53.r5 AC-3(7),
        NIST.800-53.r5 AC-5, NIST.800-53.r5 AC-6
        ECS containers should be limited to read-only access to root filesystems
        """
        name = "Ensure ECS containers are limited to read-only access to root filesystems"
        id = "CKV_AWS_336"
        supported_resources = ("aws_ecs_task_definition",)
        categories = (CheckCategories.GENERAL_SECURITY,)
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf: Dict[str, List[Any]]) -> CheckResult:
        self.evaluated_keys = ["container_definitions"]
        container_definitions = conf.get("container_definitions")
        if container_definitions and isinstance(container_definitions, list):
            containers = container_definitions[0]
            if not containers:
                return CheckResult.UNKNOWN

            if isinstance(containers, list):
                for idx, container in enumerate(containers):
                    if isinstance(container, dict) and not container.get("readonlyRootFilesystem"):
                        self.evaluated_keys = [f"container_definitions/[0]/[{idx}]/readonlyRootFilesystem"]
                        return CheckResult.FAILED
            elif isinstance(containers, dict):
                # TF plan file case
                for idx, container in enumerate(container_definitions):
                    if isinstance(container, dict) and not container.get("readonlyRootFilesystem"):
                        self.evaluated_keys = [f"container_definitions/[{idx}]/readonlyRootFilesystem"]
                        return CheckResult.FAILED

            return CheckResult.PASSED

        return CheckResult.UNKNOWN


check = ECSContainerReadOnlyRoot()
